import base64
from binascii import Error as BinError
from typing import Any, Dict, Generator, List

import pytest
from decoy import Decoy
from mock import AsyncMock, MagicMock

from opentrons.drivers.asyncio.communication.errors import NoResponse
from opentrons.drivers.asyncio.communication.serial_connection import (
    AsyncResponseSerialConnection,
)
from opentrons.drivers.flex_stacker import types
from opentrons.drivers.flex_stacker.driver import (
    DEFAULT_FS_TIMEOUT,
    FS_MOVE_TIMEOUT,
    FS_TOF_INIT_TIMEOUT,
    FS_TOF_TIMEOUT,
    FlexStackerDriver,
)
from opentrons.drivers.flex_stacker.errors import MotorStallDetected


@pytest.fixture
def connection() -> AsyncMock:
    return AsyncMock(spec=AsyncResponseSerialConnection)


@pytest.fixture
def subject(connection: AsyncMock) -> FlexStackerDriver:
    connection.send_command.return_value = ""
    connection._serial = MagicMock()
    return FlexStackerDriver(connection)


async def test_get_device_info(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get device info command"""
    connection.send_command.side_effect = [
        "M115 FW:0.0.1 HW:Opentrons-flex-stacker-a1 SerialNo:STCA120230605001",
        "M114 R:0",
    ]
    response = await subject.get_device_info()
    assert response == types.StackerInfo(
        fw="0.0.1",
        hw=types.HardwareRevision.EVT,
        sn="STCA120230605001",
        rr=0,
    )

    device_info = types.GCODE.DEVICE_INFO.build_command()
    reset_reason = types.GCODE.GET_RESET_REASON.build_command()
    connection.send_command.assert_any_call(device_info)
    connection.send_command.assert_called_with(reset_reason)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.side_effect = [
        "M115 FW:0.0.1 SerialNo:STCA120230605001",
        "M114 R:0",
    ]

    # This should raise ValueError
    with pytest.raises(ValueError):
        response = await subject.get_device_info()

    device_info = types.GCODE.DEVICE_INFO.build_command()
    connection.send_command.assert_any_call(device_info)
    # M115 response is invalid, so we dont send M114.
    connection.send_command.assert_called_once()


async def test_stop_motors(subject: FlexStackerDriver, connection: AsyncMock) -> None:
    """It should send a stop motors command"""
    connection.send_command.return_value = "M0"
    await subject.stop_motors()

    stop_motors = types.GCODE.STOP_MOTORS.build_command()
    connection.send_command.assert_any_call(stop_motors)
    connection.reset_mock()

    # This should raise ValueError
    with pytest.raises(ValueError):
        await subject.get_device_info()


async def test_get_motion_params(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get motion params command."""
    connection.send_command.return_value = "M120 M:X V:200.000 A:1500.000 D:5.000"
    response = await subject.get_motion_params(types.StackerAxis.X)
    assert response == types.MoveParams(
        acceleration=1500.0,
        max_speed=200.0,
        max_speed_discont=5.0,
    )

    command = types.GCODE.GET_MOVE_PARAMS.build_command().add_element(
        types.StackerAxis.X.name
    )
    response = await connection.send_command(command)
    connection.send_command.assert_any_call(command)
    connection.reset_mock()


async def test_set_serial_number(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a set serial number command"""
    connection.send_command.return_value = "M996"

    serial_number = "FSTA1020250119001"
    await subject.set_serial_number(serial_number)

    set_serial_number = types.GCODE.SET_SERIAL_NUMBER.build_command().add_element(
        serial_number
    )
    connection.send_command.assert_any_call(set_serial_number)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.return_value = "M9nn"
    with pytest.raises(ValueError):
        await subject.set_serial_number(serial_number)

    set_serial_number = types.GCODE.SET_SERIAL_NUMBER.build_command().add_element(
        serial_number
    )
    connection.send_command.assert_any_call(set_serial_number)
    connection.reset_mock()

    # Test invalid serial number
    with pytest.raises(ValueError):
        await subject.set_serial_number("invalid")

    connection.send_command.assert_not_called()
    connection.reset_mock()


async def test_enable_motors(subject: FlexStackerDriver, connection: AsyncMock) -> None:
    """It should send a enable motors command"""
    connection.send_command.return_value = "M17"
    await subject.enable_motors([types.StackerAxis.X])

    move_to = types.GCODE.ENABLE_MOTORS.build_command().add_element(
        types.StackerAxis.X.value
    )
    connection.send_command.assert_any_call(move_to)
    connection.reset_mock()

    # Test no arg to disable all motors
    await subject.enable_motors(list(types.StackerAxis))

    move_to = types.GCODE.ENABLE_MOTORS.build_command()
    move_to.add_element(types.StackerAxis.X.value)
    move_to.add_element(types.StackerAxis.Z.value)
    move_to.add_element(types.StackerAxis.L.value)

    connection.send_command.assert_any_call(move_to)
    connection.reset_mock()


async def test_get_limit_switch(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get limit switch command and return the boolean of one."""
    connection.send_command.return_value = "M119 XE:1 XR:0 ZE:0 ZR:1 LR:1"
    response = await subject.get_limit_switch(
        types.StackerAxis.X, types.Direction.EXTEND
    )
    assert response

    limit_switch_status = types.GCODE.GET_LIMIT_SWITCH.build_command()
    connection.send_command.assert_any_call(limit_switch_status)
    connection.reset_mock()


async def test_get_limit_switches_status(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get limit switch status and return LimitSwitchStatus."""
    connection.send_command.return_value = "M119 XE:1 XR:0 ZE:0 ZR:1 LR:1"
    response = await subject.get_limit_switches_status()
    assert response == types.LimitSwitchStatus(
        XE=True,
        XR=False,
        ZE=False,
        ZR=True,
        LR=True,
    )

    limit_switch_status = types.GCODE.GET_LIMIT_SWITCH.build_command()
    connection.send_command.assert_any_call(limit_switch_status)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.return_value = "M119 XE:b XR:0 ZE:a ZR:1 LR:n"
    with pytest.raises(ValueError):
        response = await subject.get_limit_switches_status()

    limit_switch_status = types.GCODE.GET_LIMIT_SWITCH.build_command()
    connection.send_command.assert_any_call(limit_switch_status)


async def test_get_platform_sensor(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get platform sensor command return status of specified sensor."""
    connection.send_command.return_value = "M121 E:1 R:1"
    response = await subject.get_platform_sensor(types.Direction.EXTEND)
    assert response

    platform_sensor = types.GCODE.GET_PLATFORM_SENSOR.build_command()
    connection.send_command.assert_any_call(platform_sensor)
    connection.reset_mock()


async def test_get_platform_status(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """it should send a get platform sensors status."""
    connection.send_command.return_value = "M121 E:0 R:1"
    response = await subject.get_platform_status()
    assert response == types.PlatformStatus(
        E=False,
        R=True,
    )

    platform_status = types.GCODE.GET_PLATFORM_SENSOR.build_command()
    connection.send_command.assert_any_call(platform_status)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.return_value = "M121 E:0 R:1 something"
    with pytest.raises(ValueError):
        response = await subject.get_platform_status()

    platform_status = types.GCODE.GET_PLATFORM_SENSOR.build_command()
    connection.send_command.assert_any_call(platform_status)


async def test_get_hopper_door_closed(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get door closed command."""
    connection.send_command.return_value = "M122 D:1"
    response = await subject.get_hopper_door_closed()
    assert response

    door_closed = types.GCODE.GET_DOOR_SWITCH.build_command()
    connection.send_command.assert_any_call(door_closed)
    connection.reset_mock()

    # Test door open
    connection.send_command.return_value = "M122 D:0"
    response = await subject.get_hopper_door_closed()
    assert not response

    door_closed = types.GCODE.GET_DOOR_SWITCH.build_command()
    connection.send_command.assert_any_call(door_closed)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.return_value = "M122 78gybhjk"

    with pytest.raises(ValueError):
        response = await subject.get_hopper_door_closed()

    door_closed = types.GCODE.GET_DOOR_SWITCH.build_command()
    connection.send_command.assert_any_call(door_closed)
    connection.reset_mock()


async def test_move_in_mm(subject: FlexStackerDriver, connection: AsyncMock) -> None:
    """It should send a move to command"""
    connection.send_command.return_value = "G0"
    response = await subject.move_in_mm(types.StackerAxis.X, 10)
    assert response

    move_to = types.GCODE.MOVE_TO.build_command().add_float("X", 10)
    connection.send_command.assert_any_call(move_to, timeout=FS_MOVE_TIMEOUT)
    connection.reset_mock()


@pytest.mark.parametrize(
    "command,command_args",
    [
        ("home_axis", (types.StackerAxis.X, types.Direction.EXTEND)),
        ("move_to_limit_switch", (types.StackerAxis.X, types.Direction.EXTEND)),
        ("move_in_mm", (types.StackerAxis.X, 10)),
    ],
)
async def test_move_stall(
    subject: FlexStackerDriver,
    connection: AsyncMock,
    command: str,
    command_args: list[Any],
) -> None:
    """It should handle motor stall exceptions during move commands."""
    # raise a motor stall detected exception
    connection.send_command.side_effect = MotorStallDetected(
        port="port", response="response", command="move command"
    )

    response = await subject.__getattribute__(command)(*command_args)

    connection.send_command.assert_called_once()
    # Verify we got a stall error result
    assert response == types.MoveResult.STALL_ERROR
    connection.reset_mock()


async def test_move_to_switch(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a move to switch command"""
    connection.send_command.return_value = "G5"
    axis = types.StackerAxis.X
    direction = types.Direction.EXTEND
    response = await subject.move_to_limit_switch(axis, direction)
    assert response

    move_to = types.GCODE.MOVE_TO_SWITCH.build_command().add_int(
        axis.name, direction.value
    )
    connection.send_command.assert_any_call(move_to, timeout=FS_MOVE_TIMEOUT)
    connection.reset_mock()


async def test_home_axis(subject: FlexStackerDriver, connection: AsyncMock) -> None:
    """It should send a home axis command"""
    connection.send_command.return_value = "G28"
    axis = types.StackerAxis.X
    direction = types.Direction.EXTEND
    response = await subject.home_axis(axis, direction)
    assert response

    move_to = types.GCODE.HOME_AXIS.build_command().add_int(axis.name, direction.value)
    connection.send_command.assert_any_call(move_to, timeout=FS_MOVE_TIMEOUT)
    connection.reset_mock()


async def test_set_led(subject: FlexStackerDriver, connection: AsyncMock) -> None:
    """It should send a set led command"""
    connection.send_command.return_value = "M200"
    await subject.set_led(1, types.LEDColor.RED)

    set_led = types.GCODE.SET_LED.build_command().add_float("P", 1).add_int("C", 1)
    connection.send_command.assert_any_call(set_led)
    connection.reset_mock()

    # test setting only external leds
    await subject.set_led(1, types.LEDColor.RED, external=True)

    set_led = (
        types.GCODE.SET_LED.build_command()
        .add_float("P", 1)
        .add_int("C", 1)
        .add_int("K", 1)
    )
    connection.send_command.assert_any_call(set_led)
    connection.reset_mock()


async def test_get_stallguard_threshold(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should get the stallguard threshold."""
    connection.send_command.return_value = "M911 Z:1 T:2"
    response = await subject.get_stallguard_threshold(types.StackerAxis.Z)
    assert response == types.StallGuardParams(types.StackerAxis.Z, True, 2)

    get_theshold = types.GCODE.GET_STALLGUARD_THRESHOLD.build_command().add_element(
        types.StackerAxis.Z.name
    )
    connection.send_command.assert_any_call(get_theshold)
    connection.reset_mock()


async def test_set_stallguard_threshold(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should set the stallguard threshold."""
    axis = types.StackerAxis.Z
    enable = True
    threshold = 2
    connection.send_command.return_value = "M910"
    await subject.set_stallguard_threshold(axis, enable, threshold)

    set_threshold = (
        types.GCODE.SET_STALLGUARD.build_command()
        .add_int(axis.name, int(enable))
        .add_int("T", threshold)
    )
    connection.send_command.assert_any_call(set_threshold)
    connection.reset_mock()

    # test invalid threshold
    with pytest.raises(ValueError):
        await subject.set_stallguard_threshold(axis, enable, 1000)

    connection.send_command.assert_not_called()
    connection.reset_mock()


async def test_get_motor_driver_register(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should get the motor driver register."""
    connection.send_command.return_value = "M920 Z:1 V:2"
    response = await subject.get_motor_driver_register(types.StackerAxis.Z, 1)
    assert response == 2

    get_register = types.GCODE.GET_MOTOR_DRIVER_REGISTER.build_command().add_int(
        types.StackerAxis.Z.name, 1
    )
    connection.send_command.assert_any_call(get_register)
    connection.reset_mock()


async def test_set_motor_driver_register(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should set the motor driver register."""
    connection.send_command.return_value = "M921"
    await subject.set_motor_driver_register(types.StackerAxis.Z, 1, 2)

    set_register = (
        types.GCODE.SET_MOTOR_DRIVER_REGISTER.build_command()
        .add_int(types.StackerAxis.Z.name, 1)
        .add_element(str(2))
    )
    connection.send_command.assert_any_call(set_register)
    connection.reset_mock()


async def test_enable_tof_sensor(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a enable tof sensor command."""
    connection.send_command.return_value = "M224"
    await subject.enable_tof_sensor(types.TOFSensor.X, False)

    enable_tof = types.GCODE.ENABLE_TOF_SENSOR.build_command().add_int(
        types.TOFSensor.X.name, 0
    )
    connection.send_command.assert_any_call(enable_tof, timeout=DEFAULT_FS_TIMEOUT)
    connection.reset_mock()

    # Test enable, with longer timeout
    await subject.enable_tof_sensor(types.TOFSensor.X, True)

    enable_tof = types.GCODE.ENABLE_TOF_SENSOR.build_command().add_int(
        types.TOFSensor.X.name, 1
    )
    connection.send_command.assert_any_call(enable_tof, timeout=FS_TOF_TIMEOUT)
    connection.reset_mock()


async def test_get_tof_driver_register(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should get the tof driver register."""
    connection.send_command.return_value = "M222 X:0 V:3"
    response = await subject.get_tof_driver_register(types.TOFSensor.X, 0)
    assert response == 3

    get_register = types.GCODE.GET_TOF_DRIVER_REGISTER.build_command().add_int(
        types.TOFSensor.X.name, 0
    )
    connection.send_command.assert_any_call(get_register)
    connection.reset_mock()


async def test_set_tof_driver_register(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should set the tof driver register."""
    connection.send_command.return_value = "M223"
    await subject.set_tof_driver_register(types.TOFSensor.X, 1, 1)

    set_register = (
        types.GCODE.SET_TOF_DRIVER_REGISTER.build_command()
        .add_int(types.TOFSensor.X.name, 1)
        .add_element(str(1))
    )
    connection.send_command.assert_any_call(set_register)
    connection.reset_mock()


async def test_get_tof_sensor_status(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """it should send a get tof sensors status."""
    connection.send_command.return_value = "M215 Z:1 T:2 M:3"
    response = await subject.get_tof_sensor_status(types.TOFSensor.Z)
    assert response == types.TOFSensorStatus(
        sensor=types.TOFSensor.Z,
        state=types.TOFSensorState.IDLE,
        mode=types.TOFSensorMode.MEASURE,
        ok=True,
    )

    tof_status = types.GCODE.GET_TOF_SENSOR_STATUS.build_command().add_element(
        types.TOFSensor.Z.name
    )
    connection.send_command.assert_any_call(tof_status, timeout=FS_TOF_INIT_TIMEOUT)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.return_value = "M215 Z:1 T:2 :99"
    with pytest.raises(ValueError):
        response = await subject.get_tof_sensor_status(types.TOFSensor.Z)

    tof_status = types.GCODE.GET_TOF_SENSOR_STATUS.build_command().add_element(
        types.TOFSensor.Z.name
    )
    connection.send_command.assert_any_call(tof_status, timeout=FS_TOF_INIT_TIMEOUT)


async def test_manage_tof_measurement(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """it should send a start tof measurement and receive payload info."""
    connection.send_command.return_value = "M225 X K:0 C:0 L:3840"
    response = await subject.manage_tof_measurement(types.TOFSensor.X)
    assert response == types.TOFMeasurement(
        sensor=types.TOFSensor.X,
        kind=types.MeasurementKind.HISTOGRAM,
        cancelled=False,
        total_bytes=3840,
    )

    manage_measurement = (
        types.GCODE.MANAGE_TOF_MEASUREMENT.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_int("K", types.MeasurementKind.HISTOGRAM.value)
    )
    connection.send_command.assert_any_call(manage_measurement)
    connection.reset_mock()

    # Test cancel transfer
    connection.send_command.return_value = "M225 X K:0 C:1 L:0"
    response = await subject.manage_tof_measurement(types.TOFSensor.X, start=False)
    assert response == types.TOFMeasurement(
        sensor=types.TOFSensor.X,
        kind=types.MeasurementKind.HISTOGRAM,
        cancelled=True,
        total_bytes=0,
    )

    manage_measurement = (
        types.GCODE.MANAGE_TOF_MEASUREMENT.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_int("K", types.MeasurementKind.HISTOGRAM.value)
        .add_element("C")
    )
    connection.send_command.assert_any_call(manage_measurement)
    connection.reset_mock()

    # Test invalid response
    connection.send_command.return_value = "M225 X K:0 LA0"

    # This should raise ValueError
    with pytest.raises(ValueError):
        response = await subject.manage_tof_measurement(types.TOFSensor.X)

    manage_measurement = (
        types.GCODE.MANAGE_TOF_MEASUREMENT.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_int("K", types.MeasurementKind.HISTOGRAM.value)
    )
    connection.send_command.assert_any_call(manage_measurement)
    connection.reset_mock()


async def test_get_tof_histogram_frame(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """it should send a get tof measurement and receive a frame."""
    connection.send_command.return_value = "M226 X I:30 D:gSGAAB2AAAAAAAAAAAAAA\
            AAAAAAABwAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
            AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
            AAAAAAAAAAAAAAAAAAAAAAAAAAA"
    response = await subject._get_tof_histogram_frame(types.TOFSensor.X)
    assert response == types.TOFMeasurementFrame(
        sensor=types.TOFSensor.X,
        frame_id=30,
        data=b"\x81!\x80\x00\x1d\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
    )

    get_measurement = types.GCODE.GET_TOF_MEASUREMENT.build_command().add_element(
        types.TOFSensor.X.name
    )
    connection.send_command.assert_any_call(get_measurement, retries=0)
    connection.reset_mock()

    # Test cancel transfer
    connection.send_command.return_value = "M226 X I:30 D:gSGAAB2AAAAAAAAAAAAAA\
            AAAAAAABwAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
            AAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAAA\
            AAAAAAAAAAAAAAAAAAAAAAAAAAA"
    response = await subject._get_tof_histogram_frame(types.TOFSensor.X, resend=True)
    assert response == types.TOFMeasurementFrame(
        sensor=types.TOFSensor.X,
        frame_id=30,
        data=b"\x81!\x80\x00\x1d\x80\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x07\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00",
    )

    get_measurement = (
        types.GCODE.GET_TOF_MEASUREMENT.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_element("R")
    )
    connection.send_command.assert_any_call(get_measurement, retries=0)
    connection.reset_mock()

    # Test invalid index response
    connection.send_command.return_value = "M226 X I:a D:asdsd3"

    # This should raise ValueError
    with pytest.raises(ValueError):
        response = await subject._get_tof_histogram_frame(types.TOFSensor.X)

    # Test invalid base64 frame
    connection.send_command.return_value = "M226 X I:23 D:INVALID"

    # This should raise binascii.Error
    with pytest.raises(BinError):
        response = await subject._get_tof_histogram_frame(types.TOFSensor.X)


def get_histogram_payload(frames: int) -> Generator[str, None, None]:
    """Helper to get histogram payload."""
    length = 0
    frame_id = 0
    while length < frames:
        first = bytearray(
            b"\x81\x04\x00\x0f\x00\x80\x00\x01\x00\x03\x00\x02\x02\x02\x02\x06\x03\x07\x02\x04\x03\x0b\xcc\xe5CJ\xf07I-\x9aT%\xe2\xd3\xb3\xa7\x82\x8clGKE/%.+$%\x1a\x1a\x11\x12\x10\x12\x11\x0c\x0b\t\x07\x0c\n\t\x05\x06\x0b\x0b\x05\x05\x07\x01\x0b\x05\x08\x05\x08\x03\x05\x02\x04\t\x04\x02\x02\x07\x06\x05\x06\x03\x03\x03\x07\x02\x00\x06\x03\x06\x02\x03\x05\x04\x03\x04\x02\x07\x02\x01\x05\x04\x05\x03\x02\x01\x01\x02\x02\x01\x04\x04\x04\x06\x03\x02\x04\x03\x02\x03\x01\x05\x05\x03\x02\x04\x03\x04"
        )
        rest = bytearray(
            b"\x81\x05\x80\x0e\x01\x80\x0020-*1$)213/-\x1e8\x14\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00\x00"
        )
        rest[4] = frame_id
        data = first if frame_id == 0 else rest
        encoded = base64.b64encode(data).decode("utf-8")
        yield f"M226 X I:{frame_id + 1} D:{encoded}"
        frame_id += 1
        length += 1


async def test_get_tof_histogram(
    subject: FlexStackerDriver,
    connection: AsyncMock,
    decoy: Decoy,
    histogram_bins: Dict[int, List[float]],
) -> None:
    """it should send a start and get tof measurements until full payload is transfered"""
    connection.send_command.side_effect = [
        "M215 X:1 T:2 M:3",
        "M225 X K:0 C:0 L:3840",
    ] + [p for p in get_histogram_payload(30)]
    response = await subject.get_tof_histogram(types.TOFSensor.X)
    assert response == types.TOFMeasurementResult(
        sensor=types.TOFSensor.X,
        kind=types.MeasurementKind.HISTOGRAM,
        bins=histogram_bins,
    )

    manage_measurement = (
        types.GCODE.MANAGE_TOF_MEASUREMENT.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_int("K", types.MeasurementKind.HISTOGRAM.value)
    )
    get_measurement = types.GCODE.GET_TOF_MEASUREMENT.build_command().add_element(
        types.TOFSensor.X.name
    )
    connection.send_command.assert_any_call(manage_measurement)
    connection.send_command.assert_any_call(get_measurement, retries=0)
    connection.reset_mock()

    # Test invalid frame_id
    payload = [p for p in get_histogram_payload(2)]
    connection.send_command.side_effect = (
        ["M215 X:1 T:2 M:3", "M225 X K:0 C:1 L:3840"]
        + payload
        + [payload[1], "M225 X K:0 C:1 L:0"]
    )

    # This should raise an exception
    with pytest.raises(RuntimeError):
        response = await subject.get_tof_histogram(types.TOFSensor.X)

    manage_measurement = (
        types.GCODE.MANAGE_TOF_MEASUREMENT.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_int("K", types.MeasurementKind.HISTOGRAM.value)
    )
    get_measurement = types.GCODE.GET_TOF_MEASUREMENT.build_command().add_element(
        types.TOFSensor.X.name
    )
    connection.send_command.assert_any_call(manage_measurement)
    connection.send_command.assert_any_call(get_measurement, retries=0)
    connection.reset_mock()

    # Test resend mechanism
    get_measurement = (
        types.GCODE.GET_TOF_MEASUREMENT.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_element("R")
    )
    payload = [p for p in get_histogram_payload(30)]
    connection.send_command.side_effect = (
        [
            "M215 X:1 T:2 M:3",
            "M225 X K:0 C:1 L:3840",
            payload[0],
            payload[1],
            # We raise NoResponse on frame 3 to simulate a timeout and force a resend
            NoResponse("", "Timeout"),
            # After the timeout we expect the same packet to be resent
            payload[2],
            # Then the rest of the packets
        ]
        + payload[3:]
    )

    response = await subject.get_tof_histogram(types.TOFSensor.X)

    connection.send_command.assert_any_call(manage_measurement)
    # Assert that the M226 GCODE with `R` (resend) element was sent
    connection.send_command.assert_any_call(get_measurement, retries=0)
    connection.reset_mock()


async def test_get_tof_configuration(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should get the tof sensor configuration."""
    connection.send_command.return_value = "M228 X I:14 A:110 K:4000 P:500 H:1"
    response = await subject.get_tof_configuration(types.TOFSensor.X)
    assert response == types.TOFConfiguration(
        types.TOFSensor.X,
        types.SpadMapID.SPAD_MAP_ID_14,
        types.ActiveRange.SHORT_RANGE,
        4000,
        500,
        True,
    )

    get_config = types.GCODE.GET_TOF_CONFIGURATION.build_command().add_element(
        types.TOFSensor.X.name
    )
    connection.send_command.assert_any_call(get_config)
    connection.reset_mock()


async def test_set_tof_configuration(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should set the tof sensor configuration."""
    connection.send_command.return_value = "M227"
    await subject.set_tof_configuration(
        types.TOFSensor.X,
        types.SpadMapID.SPAD_MAP_ID_1,
        types.ActiveRange.LONG_RANGE,
        3000,
        200,
        True,
    )

    set_config = (
        types.GCODE.SET_TOF_CONFIGURATION.build_command()
        .add_element(types.TOFSensor.X.name)
        .add_int("I", types.SpadMapID.SPAD_MAP_ID_1.value)
        .add_int("A", types.ActiveRange.LONG_RANGE.value)
        .add_int("K", 3000)
        .add_int("P", 200)
        .add_int("H", 1)
    )
    connection.send_command.assert_any_call(set_config)
    connection.reset_mock()


async def test_get_estop_engaged(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get estop command return boolean."""
    connection.send_command.return_value = "M112 E:1"
    response = await subject.get_estop_engaged()
    assert response

    estop = types.GCODE.GET_ESTOP_ENGAGED.build_command()
    connection.send_command.assert_any_call(estop)
    connection.reset_mock()


async def test_get_install_detected(
    subject: FlexStackerDriver, connection: AsyncMock
) -> None:
    """It should send a get install detected command and return boolean."""
    connection.send_command.return_value = "M123 I:1"
    response = await subject.get_installation_detected()
    assert response

    install_detect = types.GCODE.GET_INSTALL_DETECTED.build_command()
    connection.send_command.assert_any_call(install_detect)
    connection.reset_mock()



from math import cos, pi, sin

import numpy as np
from numpy.linalg import inv
from numpy.typing import NDArray

from opentrons.util.linal import add_z, apply_transform, solve, solve_attitude


def test_solve() -> None:
    theta = pi / 3.0  # 60 deg
    scale = 2.0

    expected = [
        (cos(0), sin(0)),
        (cos(pi / 2), sin(pi / 2)),
        (cos(pi), sin(pi)),
    ]

    actual = [
        (cos(theta) * scale + 0.5, sin(theta) * scale + 0.25),
        (cos(theta + pi / 2) * scale + 0.5, sin(theta + pi / 2) * scale + 0.25),
        (cos(theta + pi) * scale + 0.5, sin(theta + pi) * scale + 0.25),
    ]

    X = solve(expected, actual)

    expected2: NDArray[np.double] = np.array(
        [cos(theta + pi / 2) * scale + 0.5, sin(theta + pi / 2) * scale + 0.25, 1]
    )
    result = np.dot(X, np.array([[0], [1], [1]])).transpose()

    assert np.isclose(expected2, result).all()


def test_add_z() -> None:
    x = 5
    y = 10
    z = 20

    xy_array: NDArray[np.double] = np.array([[1, 0, x], [0, 1, y], [0, 0, 1]])

    expected: NDArray[np.double] = np.array(
        [[1, 0, 0, x], [0, 1, 0, y], [0, 0, 1, z], [0, 0, 0, 1]]
    )

    result = add_z(xy_array, z)
    assert (result == expected).all()


def test_apply_transform() -> None:
    x = 1
    y = 2
    z = 3

    x_delta = 0.1
    y_delta = 0.2

    e = ((1, 1, 3), (2, 2, 2), (1, 2, 1))
    a = (
        (1 + x_delta, 1 + y_delta, 1.1),
        (2 + x_delta, 2 + y_delta, 2.2),
        (1 + x_delta, 2 + y_delta, 1.1),
    )
    transform = solve_attitude(e, a)

    expected = (round(x - x_delta, 2), round(y - y_delta, 2), round(z))

    result = apply_transform(inv(transform), (1, 2, 3))
    assert np.isclose(result, expected, atol=0.1).all()



"""Test fixtures for stacker tests."""

import pytest
from decoy import Decoy

from opentrons.hardware_control.modules import FlexStacker
from opentrons.protocol_engine.execution import EquipmentHandler
from opentrons.protocol_engine.state.module_substates import (
    FlexStackerId,
)


@pytest.fixture
def stacker_id() -> FlexStackerId:
    """Get a consistent ID for a stacker."""
    return FlexStackerId("flex-stacker-id")


@pytest.fixture
def stacker_hardware(
    decoy: Decoy, equipment: EquipmentHandler, stacker_id: FlexStackerId
) -> FlexStacker:
    """Get a mocked hardware stacker."""
    hardware = decoy.mock(cls=FlexStacker)
    decoy.when(equipment.get_module_hardware_api(stacker_id)).then_return(hardware)
    return hardware



"""Test create timer command."""

import asyncio
from datetime import datetime

from decoy import Decoy, matchers

from opentrons.protocol_engine.actions import Action, ActionDispatcher, StartTaskAction
from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.create_timer import (
    CreateTimerImplementation,
    CreateTimerParams,
    CreateTimerResult,
)
from opentrons.protocol_engine.execution import RunControlHandler
from opentrons.protocol_engine.execution.task_handler import TaskHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.types.tasks import Task


async def test_create_timer_implementation(
    decoy: Decoy,
    run_control: RunControlHandler,
    real_task_handler: TaskHandler,
    action_dispatcher: ActionDispatcher,
    model_utils: ModelUtils,
) -> None:
    """It should create a task to await the duration using the RunControlHandler."""
    subject = CreateTimerImplementation(
        run_control=run_control,
        task_handler=real_task_handler,
    )
    data = CreateTimerParams(time=42.0, task_id="taskid")
    created_timestamp = datetime.now()
    decoy.when(model_utils.get_timestamp()).then_return(created_timestamp)
    decoy.when(model_utils.ensure_id("taskid")).then_return("taskid")
    task_ran = asyncio.Event()
    task: Task | None = None

    def _mock_action(action: Action) -> None:
        nonlocal task
        if isinstance(action, StartTaskAction):
            task = action.task
            task_ran.set()
        else:
            raise RuntimeError("Wrong kind of action.")

    decoy.when(
        action_dispatcher.dispatch(  # type: ignore[func-returns-value]
            StartTaskAction(
                task=Task(
                    id="taskid",
                    createdAt=created_timestamp,
                    asyncioTask=matchers.Anything(),
                )
            )
        )
    ).then_do(_mock_action)
    result = await subject.execute(data)
    assert result == SuccessData(public=CreateTimerResult(task_id="taskid", time=42.0))
    await task_ran.wait()
    assert task
    await task.asyncioTask
    decoy.verify(await run_control.wait_for_duration(seconds=42.0))



"""The versions of standard liquid classes that the Protocol API should load by default."""

from typing import TypeAlias

from opentrons.protocols.api_support.types import APIVersion

DefaultLiquidClassVersions: TypeAlias = dict[APIVersion, dict[str, int]]


# This:
#
# {
#     APIVersion(2, 100): {
#         "foo_liquid": 3,
#     },
#     APIVersion(2, 105): {
#         "foo_liquid": 7
#     }
# }
#
# Means this:
#
# apiLevels        name              Default liquid class version
# ---------------------------------------------------------------
# <2.100           foo_liquid        1
# >=2.100,<2.105   foo_liquid        3
# >=2.105          foo_liquid        7
# [any]            [anything else]   1
DEFAULT_LIQUID_CLASS_VERSIONS: DefaultLiquidClassVersions = {
    APIVersion(2, 26): {
        "ethanol_80": 2,
        "glycerol_50": 2,
        "water": 2,
    },
    APIVersion(2, 27): {
        "ethanol_80": 2,
        "glycerol_50": 2,
        "water": 3,
    },
}


def get_liquid_class_version(
    api_version: APIVersion,
    liquid_class_name: str,
) -> int:
    """Return what version of a liquid class the Protocol API should load by default."""
    default_lc_versions_newest_to_oldest = sorted(
        DEFAULT_LIQUID_CLASS_VERSIONS.items(), key=lambda kv: kv[0], reverse=True
    )
    for (
        breakpoint_api_version,
        breakpoint_liquid_class_versions,
    ) in default_lc_versions_newest_to_oldest:
        if (
            api_version >= breakpoint_api_version
            and liquid_class_name in breakpoint_liquid_class_versions
        ):
            return breakpoint_liquid_class_versions[liquid_class_name]

    return 1



import asyncio
import threading
from typing import AsyncGenerator, Iterator

import pytest

from opentrons.hardware_control import ExecutionManager
from opentrons.hardware_control.emulation.module_server import ModuleStatusClient
from opentrons.hardware_control.emulation.module_server.helpers import wait_emulators
from opentrons.hardware_control.emulation.scripts import run_app, run_smoothie
from opentrons.hardware_control.emulation.settings import (
    PipetteSettings,
    Settings,
    SmoothieSettings,
)
from opentrons.hardware_control.emulation.types import ModuleType


@pytest.fixture(scope="session")
def emulator_settings() -> Settings:
    """Emulator settings"""
    return Settings(
        smoothie=SmoothieSettings(
            left=PipetteSettings(model="p20_multi_v2.0", id="P3HMV202020041605"),
            right=PipetteSettings(model="p20_single_v2.0", id="P20SV202020070101"),
        ),
    )


@pytest.fixture(scope="session")
def emulation_app(emulator_settings: Settings) -> Iterator[None]:
    """Run the emulators"""
    modules = [
        ModuleType.Magnetic,
        ModuleType.Temperature,
        ModuleType.Thermocycler,
        ModuleType.Heatershaker,
    ]
    in_thread_loop = asyncio.new_event_loop()

    def _run_app(thread_loop: asyncio.AbstractEventLoop) -> None:
        asyncio.set_event_loop(thread_loop)

        async def _main_task() -> None:
            await asyncio.gather(
                run_smoothie.run(emulator_settings),
                run_app.run(emulator_settings, modules=[m.value for m in modules]),
            )

        try:
            thread_loop.run_until_complete(_main_task())
        except Exception:
            # this exception is from stopping the loop in the way that we will when
            # the fixture closes (by just, well, stopping the loop, and then cleaning
            # stuff up later). We can ignore it.
            pass

        thread_loop.run_until_complete(thread_loop.shutdown_asyncgens())
        thread_loop.close()

    async def _wait_ready() -> None:
        c = await ModuleStatusClient.connect(
            host="localhost",
            port=emulator_settings.module_server.port,
            interval_seconds=0.1,
        )
        await wait_emulators(client=c, modules=modules, timeout=5)
        c.close()

    def _run_wait_ready() -> None:
        asyncio.run(_wait_ready())

    # these threads are daemonized to make ctrl-c behavior a little better
    # Start the emulator thread.
    t = threading.Thread(target=_run_app, args=(in_thread_loop,))
    t.daemon = True
    t.start()

    # Start the wait for emulator ready thread and wait for it to terminate.
    ready_proc = threading.Thread(target=_run_wait_ready)
    ready_proc.daemon = True
    ready_proc.start()
    ready_proc.join()

    yield
    in_thread_loop.call_soon_threadsafe(in_thread_loop.stop)
    t.join()


@pytest.fixture
async def execution_manager() -> AsyncGenerator[ExecutionManager, None]:
    em = ExecutionManager()
    yield em
    await em.cancel()


@pytest.fixture
def poll_interval_seconds() -> float:
    """The polling interval used for the module tests.

    If too fast, tests may fail due to stale data in the serial buffers.
    If too slow, tests will take too long and may time out.
    """
    return 0.01



"""Unit tests for the `get_union_elements` module."""

import typing

import pytest

from opentrons.util.get_union_elements import get_union_elements


@pytest.mark.parametrize(
    ("input", "expected_output"),
    [
        (typing.Union[int, str, float], (int, str, float)),
        (
            typing.Annotated[typing.Union[int, str, float], "top-level annotation"],
            (int, str, float),  # The Annotated should be stripped.
        ),
        (
            typing.Annotated[
                typing.Union[int, typing.Annotated[str, "element annotation"], float],
                "top-level annotation",
            ],
            # The top-level Annotated should be stripped, but the element-level
            # Annotated should be left alone.
            (int, typing.Annotated[str, "element annotation"], float),
        ),
    ],
)
def test_get_union_elemements(
    input: typing.Any, expected_output: tuple[typing.Any, ...]
) -> None:
    """Check valid inputs to `get_union_elements()`."""
    assert get_union_elements(input) == expected_output


@pytest.mark.parametrize(
    "invalid_input",
    [
        "foo",  # not a type
        str,  # not a union
    ],
)
def test_get_union_elemements_raises_on_invalid_input(
    invalid_input: typing.Any,
) -> None:
    """Make sure invalid `get_union_elements()` inputs always raise a TypeError."""
    with pytest.raises(TypeError):
        get_union_elements(invalid_input)



"""Command precondition state and store resource."""

from dataclasses import dataclass

from ..actions import Action
from ._abstract_store import HandlesActions, HasState
from opentrons.protocol_engine.actions.get_state_update import get_state_updates
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.types import CommandPreconditions, PreconditionTypes


@dataclass
class CommandPreconditionState:
    """State of Engine command precondition references."""

    preconditions: CommandPreconditions


class CommandPreconditionStore(HasState[CommandPreconditionState], HandlesActions):
    """Command Precondition container."""

    _state: CommandPreconditionState

    def __init__(self) -> None:
        """Initialize a Command Precondition store and its state."""
        self._state = CommandPreconditionState(
            preconditions=CommandPreconditions(isCameraUsed=False)
        )

    def handle_action(self, action: Action) -> None:
        """Modify state in reaction to an action."""
        for state_update in get_state_updates(action):
            self._handle_state_update(state_update)

    def _handle_state_update(self, state_update: update_types.StateUpdate) -> None:
        if state_update.precondition_update != update_types.NO_CHANGE:
            for key in state_update.precondition_update.preconditions:
                if key == PreconditionTypes.IS_CAMERA_USED:
                    self._state.preconditions.isCameraUsed = (
                        state_update.precondition_update.preconditions[key]
                    )


class CommandPreconditionView:
    """Read-only engine created Command Precondition state view."""

    _state: CommandPreconditionState

    def __init__(self, state: CommandPreconditionState) -> None:
        """Initialize the view of Command Precondition state.

        Arguments:
            state: Command precondition dataclass used for tracking preconditions used during a protocol.
        """
        self._state = state

    def get_precondition(self) -> CommandPreconditions:
        """Get the Command Preconditions currently set by a protocol."""
        return self._state.preconditions



"""Command models for heating a Thermocycler's lid."""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Optional

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema
from typing_extensions import Literal, Type

from ...errors.error_occurrence import ErrorOccurrence
from ..command import AbstractCommandImpl, BaseCommand, BaseCommandCreate, SuccessData

if TYPE_CHECKING:
    from opentrons.protocol_engine.execution import EquipmentHandler, TaskHandler
    from opentrons.protocol_engine.state.state import StateView

SetTargetLidTemperatureCommandType = Literal["thermocycler/setTargetLidTemperature"]


def _remove_default(s: dict[str, Any]) -> None:
    s.pop("default", None)


class SetTargetLidTemperatureParams(BaseModel):
    """Input parameters to  to set a Thermocycler's target lid temperature."""

    moduleId: str = Field(..., description="Unique ID of the Thermocycler Module.")
    celsius: float = Field(..., description="Target temperature in °C.")
    taskId: str | SkipJsonSchema[None] = Field(
        None,
        description="Id for the background task that manages the temperature.",
        json_schema_extra=_remove_default,
    )


class SetTargetLidTemperatureResult(BaseModel):
    """Result data from setting a Thermocycler's target lid temperature."""

    targetLidTemperature: float = Field(
        ...,
        description="The target lid temperature that was set after validation.",
    )
    taskId: str | SkipJsonSchema[None] = Field(
        None,
        description="The task id for the setTargetBlockTemperature",
        json_schema_extra=_remove_default,
    )


class SetTargetLidTemperatureImpl(
    AbstractCommandImpl[
        SetTargetLidTemperatureParams,
        SuccessData[SetTargetLidTemperatureResult],
    ]
):
    """Execution implementation of a Thermocycler's  to set lid temperature command."""

    def __init__(
        self,
        state_view: StateView,
        equipment: EquipmentHandler,
        task_handler: TaskHandler,
        **unused_dependencies: object,
    ) -> None:
        self._state_view = state_view
        self._equipment = equipment
        self._task_handler = task_handler

    async def execute(
        self,
        params: SetTargetLidTemperatureParams,
    ) -> SuccessData[SetTargetLidTemperatureResult]:
        """To set a Thermocycler's target lid temperature."""
        thermocycler_state = self._state_view.modules.get_thermocycler_module_substate(
            params.moduleId
        )
        target_temperature = thermocycler_state.validate_target_lid_temperature(
            params.celsius
        )
        thermocycler_hardware = self._equipment.get_module_hardware_api(
            thermocycler_state.module_id
        )

        async def set_target_lid_temperature(task_handler: TaskHandler) -> None:
            if thermocycler_hardware is not None:
                async with task_handler.synchronize_cancel_latest(
                    thermocycler_state.module_id + "-lid"
                ):
                    await thermocycler_hardware.set_target_lid_temperature(
                        target_temperature
                    )
                    await thermocycler_hardware.wait_for_lid_target()

        task = await self._task_handler.create_task(
            task_function=set_target_lid_temperature, id=params.taskId
        )
        return SuccessData(
            public=SetTargetLidTemperatureResult(
                targetLidTemperature=target_temperature, taskId=task.id
            ),
        )


class SetTargetLidTemperature(
    BaseCommand[
        SetTargetLidTemperatureParams,
        SetTargetLidTemperatureResult,
        ErrorOccurrence,
    ]
):
    """A command to  to set a Thermocycler's target lid temperature."""

    commandType: SetTargetLidTemperatureCommandType = (
        "thermocycler/setTargetLidTemperature"
    )
    params: SetTargetLidTemperatureParams
    result: Optional[SetTargetLidTemperatureResult] = None

    _ImplementationCls: Type[SetTargetLidTemperatureImpl] = SetTargetLidTemperatureImpl


class SetTargetLidTemperatureCreate(BaseCommandCreate[SetTargetLidTemperatureParams]):
    """A request to create a Thermocycler's set lid temperature command."""

    commandType: SetTargetLidTemperatureCommandType = (
        "thermocycler/setTargetLidTemperature"
    )
    params: SetTargetLidTemperatureParams

    _CommandCls: Type[SetTargetLidTemperature] = SetTargetLidTemperature



"""Tests for Protocol API absorbance plate reader contexts."""

import pytest
from decoy import Decoy

from opentrons.legacy_broker import LegacyBroker
from opentrons.protocol_api import AbsorbanceReaderContext
from opentrons.protocol_api.core.common import AbsorbanceReaderCore, ProtocolCore
from opentrons.protocol_api.core.core_map import LoadedCoreMap
from opentrons.protocols.api_support.types import APIVersion


@pytest.fixture
def mock_core(decoy: Decoy) -> AbsorbanceReaderCore:
    """Get a mock module implementation core."""
    return decoy.mock(cls=AbsorbanceReaderCore)


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol implementation core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def mock_core_map(decoy: Decoy) -> LoadedCoreMap:
    """Get a mock LoadedCoreMap."""
    return decoy.mock(cls=LoadedCoreMap)


@pytest.fixture
def mock_broker(decoy: Decoy) -> LegacyBroker:
    """Get a mock command message broker."""
    return decoy.mock(cls=LegacyBroker)


@pytest.fixture
def api_version() -> APIVersion:
    """Get an API version to apply to the interface."""
    return APIVersion(2, 21)


@pytest.fixture
def subject(
    api_version: APIVersion,
    mock_core: AbsorbanceReaderCore,
    mock_protocol_core: ProtocolCore,
    mock_core_map: LoadedCoreMap,
    mock_broker: LegacyBroker,
) -> AbsorbanceReaderContext:
    """Get an absorbance reader context with its dependencies mocked out."""
    return AbsorbanceReaderContext(
        core=mock_core,
        protocol_core=mock_protocol_core,
        core_map=mock_core_map,
        broker=mock_broker,
        api_version=api_version,
    )


def test_get_serial_number(
    decoy: Decoy, mock_core: AbsorbanceReaderCore, subject: AbsorbanceReaderContext
) -> None:
    """It should get the serial number from the core."""
    decoy.when(mock_core.get_serial_number()).then_return("12345")
    result = subject.serial_number
    assert result == "12345"



from typing import NamedTuple, Optional, Union

from . import validation
from ._types import PipetteActionTypes, PlungerPositionTypes
from .core.common import ProtocolCore, RobotCore
from .labware import Labware
from .module_contexts import ModuleContext
from opentrons.hardware_control import SyncHardwareAPI
from opentrons.legacy_broker import LegacyBroker
from opentrons.legacy_commands import publisher
from opentrons.legacy_commands import robot_commands as cmds
from opentrons.protocols.api_support.types import APIVersion
from opentrons.protocols.api_support.util import requires_version
from opentrons.types import (
    AxisMapType,
    AxisType,
    DeckLocation,
    Location,
    Mount,
    Point,
    StringAxisMap,
)


class HardwareManager(NamedTuple):
    """Back. compat. wrapper for a removed class called `HardwareManager`.

    This interface will not be present in PAPIv3.
    """

    hardware: SyncHardwareAPI


class RobotContext(publisher.CommandPublisher):
    """
    A context for the movement system of the robot.

    The RobotContext class provides the objects, attributes, and methods that allow
    you to control robot motor axes individually.

    Its methods can command the robot to perform an action, like moving to an absolute
    position, controlling the gripper jaw, or moving individual pipette motors.

    Objects in this class should not be instantiated directly. Instead, instances are
    returned by [`ProtocolContext.robot()`][opentrons.protocol_api.ProtocolContext.robot].

    *New in version 2.22*
    """

    def __init__(
        self,
        core: RobotCore,
        protocol_core: ProtocolCore,
        api_version: APIVersion,
        broker: Optional[LegacyBroker] = None,
    ) -> None:
        super().__init__(broker)
        self._hardware = HardwareManager(hardware=protocol_core.get_hardware())
        self._core = core
        self._protocol_core = protocol_core
        self._api_version = api_version

    @property
    @requires_version(2, 22)
    def api_version(self) -> APIVersion:
        return self._api_version

    @property
    def hardware(self) -> HardwareManager:
        # TODO this hardware attribute should be deprecated
        # in version 3.0+ as we will only support exposed robot
        # context commands.
        return self._hardware

    @requires_version(2, 22)
    def move_to(
        self,
        mount: Union[Mount, str],
        destination: Location,
        speed: Optional[float] = None,
    ) -> None:
        """
        Move a specified mount to a location on the deck.

        Args:
            mount: The mount of the instrument you wish to move. This can either be
                an instance of [`Mount`][opentrons.types.Mount] or one of the strings
                `"left"`, `"right"`, `"extension"`, `"gripper"`. Note that the gripper
                mount can be referred to either as `"extension"` or `"gripper"`.
            destination: Any location on the deck, specified as:

                - A slot, like `"A1"`.
                - A defined location, like labware in a deck slot.
                - An absolute location, like a point `{x=10, y=10, z=10}` or a deck
                location and point (`"A1"` + point `{x=10, y=10, z=10}`).
            speed: The absolute speed in mm/s.
        """
        mount = validation.ensure_instrument_mount(mount)
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.move_to(
                # This needs to be called from protocol context and not the command for import loop reasons
                mount=mount,
                location=destination,
                speed=speed,
            ),
        ):
            self._core.move_to(mount, destination.point, speed)

    @requires_version(2, 22)
    def move_axes_to(
        self,
        axis_map: Union[AxisMapType, StringAxisMap],
        critical_point: Optional[Union[AxisMapType, StringAxisMap]] = None,
        speed: Optional[float] = None,
    ) -> None:
        """
        Move a set of axes to an absolute position on the deck.

        Args:
            axis_map (dict): A dictionary mapping axes to an absolute position on the
                deck in mm.
            critical_point (Optional[dict]): The critical point, or specific point on
                the object being moved, to move the axes with. It should only specify
                the gantry axes (i.e. `x`, `y`, `z`). When you specify a critical
                point, you're specifying the object on the gantry to be moved. If not
                specified, the critical point defaults to the center of the carriage
                attached to the gantry.
            speed (Optional[float]): The maximum speed with which to move all axes in mm/s.

        """
        instrument_on_left = self._core.get_pipette_type_from_engine(Mount.LEFT)
        is_96_channel = validation.is_pipette_96_channel(instrument_on_left)
        axis_map = validation.ensure_axis_map_type(
            axis_map, self._protocol_core.robot_type, is_96_channel
        )
        if critical_point:
            critical_point = validation.ensure_axis_map_type(
                critical_point, self._protocol_core.robot_type, is_96_channel
            )
            validation.ensure_only_gantry_axis_map_type(
                critical_point, self._protocol_core.robot_type
            )
        else:
            critical_point = None
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.move_axis_to(
                # This needs to be called from protocol context and not the command for import loop reasons
                axis_map=axis_map,
                speed=speed,
            ),
        ):
            self._core.move_axes_to(axis_map, critical_point, speed)

    @requires_version(2, 22)
    def move_axes_relative(
        self,
        axis_map: Union[AxisMapType, StringAxisMap],
        speed: Optional[float] = None,
    ) -> None:
        """
        Move a set of axes to a relative position on the deck.

        Args:
            axis_map (dict): A dictionary mapping axes to relative movements
                from the current position in mm.
            speed (float, optional): The maximum speed with which to move all
                axes in mm/s.
        """
        instrument_on_left = self._core.get_pipette_type_from_engine(Mount.LEFT)
        is_96_channel = validation.is_pipette_96_channel(instrument_on_left)

        axis_map = validation.ensure_axis_map_type(
            axis_map, self._protocol_core.robot_type, is_96_channel
        )
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.move_axis_relative(
                # This needs to be called from protocol context and not the command for import loop reasons
                axis_map=axis_map,
                speed=speed,
            ),
        ):
            self._core.move_axes_relative(axis_map, speed)

    def close_gripper_jaw(self, force: Optional[float] = None) -> None:
        """Closes the Flex Gripper jaws with a specified force.

        Args:
            force: Force with which to close the gripper jaws in newtons.
        """
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.close_gripper(
                force=force,
            ),
        ):
            self._core.close_gripper(force)

    def open_gripper_jaw(self) -> None:
        """Fully opens the Flex Gripper jaws."""
        with publisher.publish_context(
            broker=self.broker,
            command=cmds.open_gripper(),
        ):
            self._core.release_grip()

    def axis_coordinates_for(
        self,
        mount: Union[Mount, str],
        location: Union[Location, ModuleContext, DeckLocation],
    ) -> AxisMapType:
        """
        Build an axis map from a location to provide to
        either [`RobotContext.move_axes_to()`][opentrons.protocol_api.RobotContext.move_axes_to] or
        [`RobotContext.move_axes_relative()`][opentrons.protocol_api.RobotContext.move_axes_relative].
        You must provide only one of either a location, slot, or module to build
        the axis map.

        Args:
            mount (Union[types.Mount, str]): The mount of the instrument you wish to
                create an axis map for. This can either be an instance of
                [`Mount`][opentrons.types.Mount] or one of the strings `"left"`,
                `"right"`, `"extension"`, `"gripper"`. Note that the gripper mount can
                be referred to either as `"extension"` or `"gripper"`.
            location (Union[Well, ModuleContext, DeckLocation, OffDeckType]): Any
                location on the deck, specified as:

                - A deck location, like slot `"A1"`.
                - A defined location, like a module on the deck.
        """
        mount = validation.ensure_instrument_mount(mount)

        mount_axis = AxisType.axis_for_mount(mount)
        if location:
            loc: Union[Point, Labware, None]
            if isinstance(location, ModuleContext):
                loc = location.labware
                if not loc:
                    raise ValueError(f"There must be a labware on {location}")
                top_of_labware = loc.wells()[0].top()
                loc = top_of_labware.point
                return {mount_axis: loc.z, AxisType.X: loc.x, AxisType.Y: loc.y}
            elif isinstance(location, (int, str)) and not isinstance(
                location, Location
            ):
                slot_name = validation.ensure_and_convert_deck_slot(
                    location,
                    api_version=self._api_version,
                    robot_type=self._protocol_core.robot_type,
                )
                loc = self._protocol_core.get_slot_center(slot_name)
                return {mount_axis: loc.z, AxisType.X: loc.x, AxisType.Y: loc.y}
            elif isinstance(location, Location):
                assert isinstance(location, Location)
                loc = location.point
                return {mount_axis: loc.z, AxisType.X: loc.x, AxisType.Y: loc.y}
            else:
                raise ValueError(
                    "Location parameter must be a Module, Deck Location, or Location type."
                )
        else:
            raise TypeError("You must specify a location to move to.")

    def plunger_coordinates_for_volume(
        self, mount: Union[Mount, str], volume: float, action: PipetteActionTypes
    ) -> AxisMapType:
        """
        Build an axis map to move a pipette plunger motor to complete liquid handling
        actions.

        Args:
            mount: The left or right instrument mount the pipette is attached to.
            volume: A volume to convert to an axis map for linear plunger displacement.
            action: Choose to `aspirate` or `dispense`.
        """
        pipette_name = self._core.get_pipette_type_from_engine(mount)
        if not pipette_name:
            raise ValueError(
                f"Expected a pipette to be attached to provided mount {mount}"
            )
        mount = validation.ensure_mount_for_pipette(mount, pipette_name)
        pipette_axis = AxisType.plunger_axis_for_mount(mount)

        pipette_position = self._core.get_plunger_position_from_volume(
            mount, volume, action, self._protocol_core.robot_type
        )
        return {pipette_axis: pipette_position}

    def plunger_coordinates_for_named_position(
        self, mount: Union[Mount, str], position_name: PlungerPositionTypes
    ) -> AxisMapType:
        """
        Build an axis map to move a pipette plunger motor to a named position.

        Args:
            position_name: A named position to move the pipette plunger to. Choose from
                `top`, `bottom`, `blowout`, or `drop` plunger positions.

        """
        pipette_name = self._core.get_pipette_type_from_engine(mount)
        if not pipette_name:
            raise ValueError(
                f"Expected a pipette to be attached to provided mount {mount}"
            )
        mount = validation.ensure_mount_for_pipette(mount, pipette_name)
        pipette_axis = AxisType.plunger_axis_for_mount(mount)
        pipette_position = self._core.get_plunger_position_from_name(
            mount, position_name
        )
        return {pipette_axis: pipette_position}

    def build_axis_map(self, axis_map: StringAxisMap) -> AxisMapType:
        """
        Take in a [`StringAxisMap`][opentrons.types.StringAxisMap] and
        output an axis map.

        The `StringAxisMap` is allowed to contain any of the following strings:
        `"x"`, `"y"`, `"z_l"`, `"z_r"`, `"z_g"`, `"q"`.

        An example of a valid axis map could be:

            {"x": 1, "y": 2} or {"Z_L": 100}

        Note that capitalization does not matter.
        """
        instrument_on_left = self._core.get_pipette_type_from_engine(Mount.LEFT)
        is_96_channel = validation.is_pipette_96_channel(instrument_on_left)

        return validation.ensure_axis_map_type(
            axis_map, self._protocol_core.robot_type, is_96_channel
        )



"""Protocol engine types to do with modules."""

from __future__ import annotations

from dataclasses import dataclass
from enum import Enum
from typing import (
    Any,
    Dict,
    List,
    Literal,
    NamedTuple,
    Optional,
    TypeGuard,
)

from pydantic import BaseModel, Field
from pydantic.json_schema import SkipJsonSchema

from opentrons_shared_data.labware.labware_definition import Extents, LabwareDefinition
from opentrons_shared_data.labware.types import LocatingFeatures
from opentrons_shared_data.util import StrEnum

from .labware_movement import LabwareMovementOffsetData
from .labware_offset_vector import LabwareOffsetVector
from .location import DeckSlotLocation
from opentrons.hardware_control.modules import (
    ModuleModel as HardwareModuleModel,
)
from opentrons.hardware_control.modules import (
    ModuleType as ModuleType,
)


# TODO(mc, 2022-01-18): use opentrons_shared_data.module.types.ModuleModel
class ModuleModel(StrEnum):
    """All available modules' models."""

    TEMPERATURE_MODULE_V1 = "temperatureModuleV1"
    TEMPERATURE_MODULE_V2 = "temperatureModuleV2"
    MAGNETIC_MODULE_V1 = "magneticModuleV1"
    MAGNETIC_MODULE_V2 = "magneticModuleV2"
    THERMOCYCLER_MODULE_V1 = "thermocyclerModuleV1"
    THERMOCYCLER_MODULE_V2 = "thermocyclerModuleV2"
    HEATER_SHAKER_MODULE_V1 = "heaterShakerModuleV1"
    MAGNETIC_BLOCK_V1 = "magneticBlockV1"
    ABSORBANCE_READER_V1 = "absorbanceReaderV1"
    FLEX_STACKER_MODULE_V1 = "flexStackerModuleV1"
    VACUUM_MODULE_V1 = "vacuumModuleMilliporeV1"

    @classmethod
    def from_hardware(cls, hardware_model: HardwareModuleModel) -> "ModuleModel":
        """Convert from the hardware model representation."""
        return cls(hardware_model.value)

    def as_type(self) -> ModuleType:
        """Get the ModuleType of this model."""
        if ModuleModel.is_temperature_module_model(self):
            return ModuleType.TEMPERATURE
        elif ModuleModel.is_magnetic_module_model(self):
            return ModuleType.MAGNETIC
        elif ModuleModel.is_thermocycler_module_model(self):
            return ModuleType.THERMOCYCLER
        elif ModuleModel.is_heater_shaker_module_model(self):
            return ModuleType.HEATER_SHAKER
        elif ModuleModel.is_magnetic_block(self):
            return ModuleType.MAGNETIC_BLOCK
        elif ModuleModel.is_absorbance_reader(self):
            return ModuleType.ABSORBANCE_READER
        elif ModuleModel.is_flex_stacker(self):
            return ModuleType.FLEX_STACKER
        elif ModuleModel.is_vacuum_module(self):
            return ModuleType.VACUUM_MODULE

        assert False, f"Invalid ModuleModel {self}"

    @classmethod
    def is_temperature_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[TemperatureModuleModel]:
        """Whether a given model is a Temperature Module."""
        return model in [cls.TEMPERATURE_MODULE_V1, cls.TEMPERATURE_MODULE_V2]

    @classmethod
    def is_magnetic_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[MagneticModuleModel]:
        """Whether a given model is a Magnetic Module."""
        return model in [cls.MAGNETIC_MODULE_V1, cls.MAGNETIC_MODULE_V2]

    @classmethod
    def is_thermocycler_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[ThermocyclerModuleModel]:
        """Whether a given model is a Thermocycler Module."""
        return model in [cls.THERMOCYCLER_MODULE_V1, cls.THERMOCYCLER_MODULE_V2]

    @classmethod
    def is_heater_shaker_module_model(
        cls, model: ModuleModel
    ) -> TypeGuard[HeaterShakerModuleModel]:
        """Whether a given model is a Heater-Shaker Module."""
        return model == cls.HEATER_SHAKER_MODULE_V1

    @classmethod
    def is_magnetic_block(cls, model: ModuleModel) -> TypeGuard[MagneticBlockModel]:
        """Whether a given model is a Magnetic block."""
        return model == cls.MAGNETIC_BLOCK_V1

    @classmethod
    def is_absorbance_reader(
        cls, model: ModuleModel
    ) -> TypeGuard[AbsorbanceReaderModel]:
        """Whether a given model is an Absorbance Plate Reader."""
        return model == cls.ABSORBANCE_READER_V1

    @classmethod
    def is_flex_stacker(cls, model: ModuleModel) -> TypeGuard[FlexStackerModuleModel]:
        """Whether a given model is a Flex Stacker.."""
        return model == cls.FLEX_STACKER_MODULE_V1

    @classmethod
    def is_vacuum_module(cls, model: ModuleModel) -> TypeGuard[VacuumModuleModel]:
        """Whether a given model is a Vacuum Module."""
        return model == cls.VACUUM_MODULE_V1


TemperatureModuleModel = Literal[
    ModuleModel.TEMPERATURE_MODULE_V1, ModuleModel.TEMPERATURE_MODULE_V2
]
MagneticModuleModel = Literal[
    ModuleModel.MAGNETIC_MODULE_V1, ModuleModel.MAGNETIC_MODULE_V2
]
ThermocyclerModuleModel = Literal[
    ModuleModel.THERMOCYCLER_MODULE_V1, ModuleModel.THERMOCYCLER_MODULE_V2
]
HeaterShakerModuleModel = Literal[ModuleModel.HEATER_SHAKER_MODULE_V1]
MagneticBlockModel = Literal[ModuleModel.MAGNETIC_BLOCK_V1]
AbsorbanceReaderModel = Literal[ModuleModel.ABSORBANCE_READER_V1]
FlexStackerModuleModel = Literal[ModuleModel.FLEX_STACKER_MODULE_V1]
VacuumModuleModel = Literal[ModuleModel.VACUUM_MODULE_V1]


class ModuleDimensions(BaseModel):
    """Dimension type for modules."""

    bareOverallHeight: float
    overLabwareHeight: float
    labwareInterfaceXDimension: Optional[float] = None
    labwareInterfaceYDimension: Optional[float] = None
    lidHeight: Optional[float] = None
    maxStackerFillHeight: Optional[float] = None
    maxStackerRetrievableHeight: Optional[float] = None


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class ModuleCalibrationPoint(BaseModel):
    """Calibration Point type for module definition."""

    x: float
    y: float
    z: float


# TODO(mm, 2023-04-13): Move to shared-data, so this binding can be maintained alongside the JSON
# schema that it's sourced from. We already do that for labware definitions and JSON protocols.


# See underlying JSON schema for documentation.
class ModuleDefinition(BaseModel):
    """A module definition conforming to module definition schema v3."""

    # Note: This field is misleading.
    #
    # This class only models v3 definitions ("module/schemas/3"), not v2 ("module/schemas/2").
    # labwareOffset is required to have a z-component, for example.
    #
    # When parsing from a schema v3 JSON definition into this model,
    # the definition's `"$otSharedSchema": "module/schemas/3"` field will be thrown away
    # because it has a dollar sign, which doesn't match this field.
    # Then, this field will default to "module/schemas/2", because no value was provided.
    #
    # We should fix this field once Jira RSS-221 is resolved. RSS-221 makes it difficult to fix
    # because robot-server has been storing and loading these bad fields in its database.
    otSharedSchema: str = Field("module/schemas/2", description="The current schema.")

    moduleType: ModuleType = Field(...)

    model: ModuleModel = Field(...)

    labwareOffset: LabwareOffsetVector = Field(...)

    dimensions: ModuleDimensions = Field(...)

    calibrationPoint: ModuleCalibrationPoint = Field(
        ...,
    )

    displayName: str = Field(...)

    quirks: List[str] = Field(...)

    # In releases prior to https://github.com/Opentrons/opentrons/pull/11873 (v6.3.0),
    # the matrices in slotTransforms were 3x3.
    # After, they are 4x4, even though there was no schema version bump.
    #
    # Because old objects of this class, with the 3x3 matrices, were stored in robot-server's
    # database, this field needs to stay typed loosely enough to support both sizes.
    # We can fix this once Jira RSS-221 is resolved.
    slotTransforms: Dict[str, Any] = Field(
        ...,
    )

    orientation: Dict[str, str] = Field(
        ...,
    )

    compatibleWith: List[ModuleModel] = Field(
        ...,
    )
    gripperOffsets: Optional[Dict[str, LabwareMovementOffsetData]] = Field(
        default_factory=lambda: {},
    )

    features: LocatingFeatures = Field(
        ...,
    )

    extents: Extents = Field(
        ...,
    )


class LoadedModule(BaseModel):
    """A module that has been loaded."""

    id: str
    model: ModuleModel
    location: Optional[DeckSlotLocation] = None
    serialNumber: Optional[str] = None


class SpeedRange(NamedTuple):
    """Minimum and maximum allowed speeds for a shaking module."""

    min: int
    max: int


class TemperatureRange(NamedTuple):
    """Minimum and maximum allowed temperatures for a heating module."""

    min: float
    max: float


class HeaterShakerLatchStatus(Enum):
    """Heater-Shaker latch status for determining pipette and labware movement errors."""

    CLOSED = "closed"
    OPEN = "open"
    UNKNOWN = "unknown"


@dataclass(frozen=True)
class HeaterShakerMovementRestrictors:
    """Shaking status, latch status and slot location for determining movement restrictions."""

    plate_shaking: bool
    latch_status: HeaterShakerLatchStatus
    deck_slot: int


ABSMeasureMode = Literal["single", "multi"]


# TODO(mm, 2022-11-07): Deduplicate with Vec3f.
class ModuleOffsetVector(BaseModel):
    """Offset, in deck coordinates, from nominal to actual position of labware on a module."""

    x: float
    y: float
    z: float


@dataclass
class ModuleOffsetData:
    """Module calibration offset data."""

    moduleOffsetVector: ModuleOffsetVector
    location: DeckSlotLocation


class StackerFillEmptyStrategy(StrEnum):
    """Strategy to use for filling or emptying a stacker."""

    MANUAL_WITH_PAUSE = "manualWithPause"
    LOGICAL = "logical"


class StackerLabwareMovementStrategy(StrEnum):
    """Strategy to retrieve or store labware."""

    AUTOMATIC = "automatic"
    MANUAL = "manual"


class StackerStoredLabwareGroup(BaseModel):
    """Represents one group of labware stored in a stacker hopper."""

    primaryLabwareId: str
    adapterLabwareId: str | SkipJsonSchema[None] = None
    lidLabwareId: str | SkipJsonSchema[None] = None


@dataclass
class StackerPoolDefinition:
    """Represents an internal configuration of stored labware."""

    primaryLabwareDefinition: LabwareDefinition
    adapterLabwareDefinition: LabwareDefinition | SkipJsonSchema[None] = None
    lidLabwareDefinition: LabwareDefinition | SkipJsonSchema[None] = None


class IdentifyColor(StrEnum):
    """Module identify color."""

    WHITE = "white"
    RED = "red"
    GREEN = "green"
    BLUE = "blue"
    YELLOW = "yellow"



from datetime import datetime

from ..tasks import AbstractTaskCore
from opentrons.protocol_engine.clients import SyncClient as ProtocolEngineClient
from opentrons.protocol_engine.errors.exceptions import NoTaskFoundError


class EngineTaskCore(AbstractTaskCore):
    def __init__(
        self, task_id: str | None, engine_client: ProtocolEngineClient
    ) -> None:
        self._id = task_id
        self._engine_client = engine_client

    def get_created_at_timestamp(self) -> datetime:
        if self._id is None:
            raise NoTaskFoundError
        try:
            task = self._engine_client.state.tasks.get(self._id)
            return task.createdAt
        except NoTaskFoundError:
            raise NoTaskFoundError

    def is_done(self) -> bool:
        if self._id is None:
            raise NoTaskFoundError
        try:
            self._engine_client.state.tasks.get_finished(self._id)
            return True
        except NoTaskFoundError:
            return False

    def is_started(self) -> bool:
        if self._id is None:
            raise NoTaskFoundError
        try:
            self._engine_client.state.tasks.get_current(self._id)
            return True
        except NoTaskFoundError:
            return self.is_done()

    def get_finished_at_timestamp(self) -> datetime | None:
        if self._id is None:
            raise NoTaskFoundError
        try:
            finished_task = self._engine_client.state.tasks.get_finished(self._id)
            return finished_task.finishedAt
        except NoTaskFoundError:
            return None



"""Tests for the legacy Protocol API core implementation."""

import inspect
from typing import Any, Dict, cast

import pytest
from decoy import Decoy

from opentrons_shared_data.labware.types import LabwareDefinition2 as LabwareDefDict
from opentrons_shared_data.module.types import ModuleDefinitionV3
from opentrons_shared_data.pipette.types import PipetteNameType

from opentrons.hardware_control import SyncHardwareAPI
from opentrons.hardware_control.dev_types import PipetteDict
from opentrons.hardware_control.modules import AbstractModule
from opentrons.hardware_control.modules.types import ModuleType, TemperatureModuleModel
from opentrons.protocol_api import MAX_SUPPORTED_VERSION, OFF_DECK
from opentrons.protocol_api.core.labware import LabwareLoadParams
from opentrons.protocol_api.core.legacy import (
    legacy_module_core as mock_legacy_module_core,
)
from opentrons.protocol_api.core.legacy import (
    module_geometry as mock_module_geometry,
)
from opentrons.protocol_api.core.legacy.deck import Deck
from opentrons.protocol_api.core.legacy.labware_offset_provider import (
    AbstractLabwareOffsetProvider,
    ProvidedLabwareOffset,
)
from opentrons.protocol_api.core.legacy.legacy_instrument_core import (
    LegacyInstrumentCore,
)
from opentrons.protocol_api.core.legacy.legacy_labware_core import LegacyLabwareCore
from opentrons.protocol_api.core.legacy.legacy_module_core import LegacyModuleCore
from opentrons.protocol_api.core.legacy.legacy_protocol_core import (
    LegacyProtocolCore,
)
from opentrons.protocol_api.core.legacy.load_info import (
    InstrumentLoadInfo,
    LabwareLoadInfo,
    LoadInfo,
    ModuleLoadInfo,
)
from opentrons.protocol_api.core.legacy.module_geometry import ModuleGeometry
from opentrons.protocols import labware as mock_labware
from opentrons.protocols.api_support.util import APIVersionError
from opentrons.types import DeckSlotName, Location, Mount, Point, StagingSlotName
from opentrons.util.broker import Broker


@pytest.fixture(autouse=True)
def _mock_labware_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out opentrons.protocols.labware functions."""
    for name, func in inspect.getmembers(mock_labware, inspect.isfunction):
        monkeypatch.setattr(mock_labware, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _mock_module_geometry_module(decoy: Decoy, monkeypatch: pytest.MonkeyPatch) -> None:
    """Mock out opentrons.protocols.geometry.module_geometry functions."""
    for name, func in inspect.getmembers(mock_module_geometry, inspect.isfunction):
        monkeypatch.setattr(mock_module_geometry, name, decoy.mock(func=func))


@pytest.fixture(autouse=True)
def _mock_legacy_module_core_module(
    decoy: Decoy, monkeypatch: pytest.MonkeyPatch
) -> None:
    """Mock out opentrons.protocol_api.core.legacy.legacy_module_core functions."""
    for name, func in inspect.getmembers(mock_legacy_module_core, inspect.isfunction):
        monkeypatch.setattr(mock_legacy_module_core, name, decoy.mock(func=func))


@pytest.fixture
def mock_sync_hardware_api(decoy: Decoy) -> SyncHardwareAPI:
    """Get a mock hardware API."""
    return decoy.mock(cls=SyncHardwareAPI)


@pytest.fixture
def mock_deck(decoy: Decoy) -> Deck:
    """Get a mock Deck."""

    class _MockDeck(Dict[str, Any]): ...

    deck = _MockDeck()
    setattr(deck, "position_for", decoy.mock(name="Deck.position_for"))
    setattr(
        deck, "resolve_module_location", decoy.mock(name="Deck.resolve_module_location")
    )
    deck["12"] = decoy.mock(cls=LegacyLabwareCore)

    return cast(Deck, deck)


@pytest.fixture
def mock_labware_offset_provider(decoy: Decoy) -> AbstractLabwareOffsetProvider:
    """Get a mock offset provider core."""
    return decoy.mock(cls=AbstractLabwareOffsetProvider)


@pytest.fixture
def mock_equipment_broker(decoy: Decoy) -> Broker[LoadInfo]:
    """Get a mock equipment broker."""
    return decoy.mock(cls=Broker)


@pytest.fixture
def subject(
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_labware_offset_provider: AbstractLabwareOffsetProvider,
    mock_equipment_broker: Broker[LoadInfo],
    mock_deck: Deck,
) -> LegacyProtocolCore:
    """Get a legacy protocol implementation core with mocked out dependencies."""
    return LegacyProtocolCore(
        api_version=MAX_SUPPORTED_VERSION,
        sync_hardware=mock_sync_hardware_api,
        labware_offset_provider=mock_labware_offset_provider,
        equipment_broker=mock_equipment_broker,
        deck_layout=mock_deck,
    )


def test_load_instrument(
    decoy: Decoy,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_equipment_broker: Broker[LoadInfo],
    subject: LegacyProtocolCore,
) -> None:
    """It should load an instrument core."""
    pipette_dict = cast(
        PipetteDict,
        {
            "model": "cool-model",
            "pipette_id": "cool-serial-number",
            "default_aspirate_flow_rates": {"1.1": 22},
            "default_dispense_flow_rates": {"3.3": 44},
            "default_blow_out_flow_rates": {"5.5": 66},
        },
    )

    decoy.when(mock_sync_hardware_api.attached_instruments).then_return({})
    decoy.when(mock_sync_hardware_api.get_attached_instrument(Mount.RIGHT)).then_return(
        pipette_dict
    )

    result = subject.load_instrument(
        instrument_name=PipetteNameType.P300_SINGLE, mount=Mount.RIGHT
    )

    assert isinstance(result, LegacyInstrumentCore)

    decoy.verify(
        mock_sync_hardware_api.cache_instruments({Mount.RIGHT: "p300_single"}),
        mock_equipment_broker.publish(
            InstrumentLoadInfo(
                instrument_load_name="p300_single",
                mount=Mount.RIGHT,
                pipette_dict=pipette_dict,
            )
        ),
    )


def test_load_labware_off_deck_raises(
    subject: LegacyProtocolCore,
) -> None:
    """It should raise an api error."""
    with pytest.raises(APIVersionError):
        subject.load_labware(
            load_name="cool load name",
            location=OFF_DECK,
            label="cool label",
            namespace="cool namespace",
            version=1337,
        )


def test_load_labware_on_staging_slot_raises(
    subject: LegacyProtocolCore,
) -> None:
    """It should raise an api error when loading onto a staging slot."""
    with pytest.raises(APIVersionError):
        subject.load_labware(
            load_name="cool load name",
            location=StagingSlotName.SLOT_B4,
            label="cool label",
            namespace="cool namespace",
            version=1337,
        )


def test_load_labware(
    decoy: Decoy,
    mock_deck: Deck,
    mock_labware_offset_provider: AbstractLabwareOffsetProvider,
    mock_equipment_broker: Broker[LoadInfo],
    subject: LegacyProtocolCore,
) -> None:
    """It should load a labware core."""
    labware_definition_dict = cast(
        LabwareDefDict,
        {
            "schemaVersion": 2,
            "namespace": "super cool namespace",
            "parameters": {"loadName": "super cool load name"},
            "version": 42,
            "wells": {},
            "ordering": [],
            "dimensions": {"xDimension": 0, "yDimension": 0, "zDimension": 0},
            "cornerOffsetFromSlot": {"x": 4, "y": 5, "z": 6},
        },
    )

    decoy.when(
        mock_labware.get_labware_definition(
            load_name="cool load name",
            namespace="cool namespace",
            version=1337,
            bundled_defs=None,
            extra_defs={},
        )
    ).then_return(labware_definition_dict)

    decoy.when(mock_deck.position_for("5")).then_return(
        Location(Point(1, 2, 3), "5"),
    )

    decoy.when(
        mock_labware_offset_provider.find(
            load_params=LabwareLoadParams(
                namespace="super cool namespace",
                load_name="super cool load name",
                version=42,
            ),
            requested_module_model=None,
            deck_slot=DeckSlotName.SLOT_5,
        )
    ).then_return(ProvidedLabwareOffset(delta=Point(7, 8, 9), offset_id="offset-789"))

    result = subject.load_labware(
        load_name="cool load name",
        location=DeckSlotName.SLOT_5,
        label="cool label",
        namespace="cool namespace",
        version=1337,
    )

    assert isinstance(result, LegacyLabwareCore)
    assert mock_deck["5"] is result
    assert result.get_definition() == labware_definition_dict
    assert result.get_user_display_name() == "cool label"
    assert result.get_geometry().parent == Location(Point(1, 2, 3), "5")
    assert result.get_calibrated_offset() == Point(
        x=(1 + 4 + 7),
        y=(2 + 5 + 8),
        z=(3 + 6 + 9),
    )

    decoy.verify(
        mock_equipment_broker.publish(
            LabwareLoadInfo(
                labware_definition=labware_definition_dict,
                labware_namespace="super cool namespace",
                labware_load_name="super cool load name",
                labware_version=42,
                deck_slot=DeckSlotName.SLOT_5,
                on_module=False,
                offset_id="offset-789",
                labware_display_name="cool label",
            )
        ),
        times=1,
    )


def test_load_adapter_raises(
    subject: LegacyProtocolCore,
) -> None:
    """It should raise an API version error when trying to load an adapter."""
    with pytest.raises(APIVersionError):
        subject.load_adapter(
            load_name="cool load name",
            location=DeckSlotName.SLOT_5,
            namespace="cool namespace",
            version=1337,
        )


def test_load_labware_on_module(
    decoy: Decoy,
    mock_labware_offset_provider: AbstractLabwareOffsetProvider,
    mock_equipment_broker: Broker[LoadInfo],
    subject: LegacyProtocolCore,
) -> None:
    """It should load a labware core."""
    mock_module_core = decoy.mock(cls=LegacyModuleCore)

    labware_definition_dict = cast(
        LabwareDefDict,
        {
            "schemaVersion": 2,
            "namespace": "super cool namespace",
            "parameters": {"loadName": "super cool load name"},
            "version": 42,
            "wells": {},
            "ordering": [],
            "dimensions": {"xDimension": 0, "yDimension": 0, "zDimension": 0},
            "cornerOffsetFromSlot": {"x": 4, "y": 5, "z": 6},
        },
    )

    decoy.when(
        mock_labware.get_labware_definition(
            load_name="cool load name",
            namespace="cool namespace",
            version=1337,
            bundled_defs=None,
            extra_defs={},
        )
    ).then_return(labware_definition_dict)

    decoy.when(mock_module_core.get_requested_model()).then_return(
        TemperatureModuleModel.TEMPERATURE_V1
    )
    decoy.when(mock_module_core.get_deck_slot()).then_return(DeckSlotName.SLOT_5)
    decoy.when(mock_module_core.geometry.location).then_return(
        Location(Point(1, 2, 3), mock_module_core.geometry)
    )

    decoy.when(
        mock_labware_offset_provider.find(
            load_params=LabwareLoadParams(
                namespace="super cool namespace",
                load_name="super cool load name",
                version=42,
            ),
            requested_module_model=TemperatureModuleModel.TEMPERATURE_V1,
            deck_slot=DeckSlotName.SLOT_5,
        )
    ).then_return(ProvidedLabwareOffset(delta=Point(7, 8, 9), offset_id="offset-789"))

    result = subject.load_labware(
        load_name="cool load name",
        location=mock_module_core,
        label="cool label",
        namespace="cool namespace",
        version=1337,
    )

    assert isinstance(result, LegacyLabwareCore)
    assert result.get_calibrated_offset() == Point(
        x=(1 + 4 + 7),
        y=(2 + 5 + 8),
        z=(3 + 6 + 9),
    )

    decoy.verify(
        mock_equipment_broker.publish(
            LabwareLoadInfo(
                labware_definition=labware_definition_dict,
                labware_namespace="super cool namespace",
                labware_load_name="super cool load name",
                labware_version=42,
                deck_slot=DeckSlotName.SLOT_5,
                on_module=True,
                offset_id="offset-789",
                labware_display_name="cool label",
            )
        ),
        times=1,
    )


def test_load_labware_on_labware_raises(
    decoy: Decoy, subject: LegacyProtocolCore
) -> None:
    """It should raise an API version error when trying to load a labware onto a labware."""
    with pytest.raises(APIVersionError):
        subject.load_labware(
            load_name="cool load name",
            location=decoy.mock(cls=LegacyLabwareCore),
            label="cool label",
            namespace="cool namespace",
            version=1337,
        )


def test_load_module(
    decoy: Decoy,
    mock_deck: Deck,
    mock_sync_hardware_api: SyncHardwareAPI,
    mock_equipment_broker: Broker[LoadInfo],
    subject: LegacyProtocolCore,
) -> None:
    """It should load a module core.

    Note: this test does not fully test the the loading logic,
    including (but not limited to) module simulation logic.
    Since this code is being replaced and refactoring would be
    prohibitively expensive, this should be good enough
    given existing high-level acceptance tests.
    """
    mock_hw_mod_1 = decoy.mock(cls=AbstractModule)
    mock_hw_mod_2 = decoy.mock(cls=AbstractModule)

    defn_1 = cast(ModuleDefinitionV3, {"model": "model-1"})
    defn_2 = cast(ModuleDefinitionV3, {"model": "model-2"})

    mock_geometry = decoy.mock(cls=ModuleGeometry)
    mock_module_core = decoy.mock(cls=LegacyModuleCore)

    decoy.when(mock_hw_mod_1.model()).then_return("model-1")
    decoy.when(mock_hw_mod_2.model()).then_return("model-2")

    decoy.when(
        mock_deck.resolve_module_location(
            ModuleType.TEMPERATURE, DeckSlotName.SLOT_1.id
        )
    ).then_return(42)

    decoy.when(mock_deck.position_for(42)).then_return(Location(Point(1, 2, 3), None))

    decoy.when(mock_sync_hardware_api.attached_modules).then_return(
        [mock_hw_mod_1, mock_hw_mod_2]
    )

    decoy.when(mock_module_geometry.load_definition("model-1")).then_return(defn_1)
    decoy.when(mock_module_geometry.load_definition("model-2")).then_return(defn_2)

    decoy.when(
        mock_module_geometry.models_compatible(
            TemperatureModuleModel.TEMPERATURE_V1, defn_1
        )
    ).then_return(False)

    decoy.when(
        mock_module_geometry.models_compatible(
            TemperatureModuleModel.TEMPERATURE_V1, defn_2
        )
    ).then_return(True)

    decoy.when(
        mock_module_geometry.create_geometry(
            definition=defn_2,
            parent=Location(Point(1, 2, 3), None),
            configuration=None,
        )
    ).then_return(mock_geometry)

    decoy.when(mock_hw_mod_2.device_info).then_return({"serial": "serial-number"})

    decoy.when(mock_geometry.parent).then_return("1")
    decoy.when(mock_geometry.model).then_return(TemperatureModuleModel.TEMPERATURE_V2)

    decoy.when(
        mock_legacy_module_core.create_module_core(
            module_hardware_api=mock_hw_mod_2,
            requested_model=TemperatureModuleModel.TEMPERATURE_V1,
            geometry=mock_geometry,
            protocol_core=subject,
        )
    ).then_return(mock_module_core)

    decoy.when(mock_module_core.get_model()).then_return(
        TemperatureModuleModel.TEMPERATURE_V2
    )
    decoy.when(mock_module_core.get_serial_number()).then_return("cap'n crunch")
    decoy.when(mock_module_core.get_deck_slot()).then_return(DeckSlotName.SLOT_1)

    result = subject.load_module(
        model=TemperatureModuleModel.TEMPERATURE_V1,
        deck_slot=DeckSlotName.SLOT_1,
        configuration=None,
    )

    assert result is mock_module_core

    decoy.verify(
        mock_equipment_broker.publish(
            ModuleLoadInfo(
                requested_model=TemperatureModuleModel.TEMPERATURE_V1,
                loaded_model=TemperatureModuleModel.TEMPERATURE_V2,
                module_serial="cap'n crunch",
                deck_slot=DeckSlotName.SLOT_1,
                configuration=None,
            )
        ),
        times=1,
    )



"""Test dispense-in-place commands."""

from datetime import datetime
from typing import Optional

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import PipetteOverpressureError

from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.dispense_in_place import (
    DispenseInPlaceImplementation,
    DispenseInPlaceParams,
    DispenseInPlaceResult,
)
from opentrons.protocol_engine.commands.pipetting_common import OverpressureError
from opentrons.protocol_engine.execution import GantryMover, PipettingHandler
from opentrons.protocol_engine.resources import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import (
    CurrentAddressableArea,
    CurrentPipetteLocation,
    CurrentWell,
)
from opentrons.types import Point


@pytest.fixture
def subject(
    pipetting: PipettingHandler,
    state_view: StateView,
    gantry_mover: GantryMover,
    model_utils: ModelUtils,
) -> DispenseInPlaceImplementation:
    """Build a command implementation."""
    return DispenseInPlaceImplementation(
        pipetting=pipetting,
        state_view=state_view,
        gantry_mover=gantry_mover,
        model_utils=model_utils,
    )


@pytest.mark.parametrize(
    "location,stateupdateLabware,stateupdateWell",
    [
        (
            CurrentWell(
                pipette_id="pipette-id-abc",
                labware_id="labware-id-1",
                well_name="well-name-1",
            ),
            "labware-id-1",
            "well-name-1",
        ),
        (None, None, None),
        (CurrentAddressableArea("pipette-id-abc", "addressable-area-1"), None, None),
    ],
)
async def test_dispense_in_place_implementation(
    decoy: Decoy,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    state_view: StateView,
    subject: DispenseInPlaceImplementation,
    location: CurrentPipetteLocation | None,
    stateupdateLabware: str,
    stateupdateWell: str,
) -> None:
    """It should dispense in place."""
    data = DispenseInPlaceParams(
        pipetteId="pipette-id-abc",
        volume=123,
        flowRate=456,
    )
    decoy.when(pipetting.get_state_view()).then_return(state_view)

    decoy.when(
        await pipetting.dispense_in_place(
            pipette_id="pipette-id-abc",
            volume=123,
            flow_rate=456,
            push_out=None,
            is_full_dispense=False,
            correction_volume=0,
        )
    ).then_return(42)

    decoy.when(state_view.pipettes.get_current_location()).then_return(location)
    decoy.when(
        state_view.pipettes.get_liquid_dispensed_by_ejecting_volume(
            pipette_id="pipette-id-abc", volume=42
        )
    ).then_return(34)

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=stateupdateLabware,
            target_well_name=stateupdateWell,
            pipette_id="pipette-id-abc",
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            stateupdateLabware, stateupdateWell, "pipette-id-abc"
        )
    ).then_return(["A3", "A4"])
    decoy.when(await gantry_mover.get_position("pipette-id-abc")).then_return(
        Point(1, 2, 3)
    )

    result = await subject.execute(data)

    if isinstance(location, CurrentWell):
        assert result == SuccessData(
            public=DispenseInPlaceResult(volume=42),
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                    pipette_id="pipette-id-abc", volume=42
                ),
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id=stateupdateLabware,
                    well_names=["A3", "A4"],
                    volume_added=68,
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id-abc", ready_to_aspirate=True
                ),
            ),
        )
    else:
        assert result == SuccessData(
            public=DispenseInPlaceResult(volume=42),
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                    pipette_id="pipette-id-abc", volume=42
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id-abc", ready_to_aspirate=True
                ),
            ),
        )


@pytest.mark.parametrize(
    "asp_volume, dispense_volume, push_out, expected",
    [
        (50.0, 50.0, None, False),
        (50.0, 50.0, 0.0, True),
        (100.0, 50.0, None, True),
    ],
)
async def test_dispense_in_place_is_ready_response(
    decoy: Decoy,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    state_view: StateView,
    subject: DispenseInPlaceImplementation,
    asp_volume: float,
    dispense_volume: float,
    push_out: Optional[float],
    expected: bool,
) -> None:
    """It should dispense in place."""
    data = DispenseInPlaceParams(
        pipetteId="pipette-id-abc",
        volume=dispense_volume,
        flowRate=456,
        pushOut=push_out,
    )
    decoy.when(pipetting.get_state_view()).then_return(state_view)
    decoy.when(state_view.pipettes.get_aspirated_volume("pipette-id-abc")).then_return(
        asp_volume
    )

    decoy.when(
        await pipetting.dispense_in_place(
            pipette_id="pipette-id-abc",
            volume=dispense_volume,
            flow_rate=456,
            push_out=push_out,
            is_full_dispense=dispense_volume == asp_volume,
            correction_volume=0,
        )
    ).then_return(dispense_volume)

    decoy.when(state_view.pipettes.get_current_location()).then_return(None)
    decoy.when(
        state_view.pipettes.get_liquid_dispensed_by_ejecting_volume(
            pipette_id="pipette-id-abc", volume=dispense_volume
        )
    ).then_return(dispense_volume)

    decoy.when(await gantry_mover.get_position("pipette-id-abc")).then_return(
        Point(1, 2, 3)
    )

    result = await subject.execute(data)

    assert result == SuccessData(
        public=DispenseInPlaceResult(volume=dispense_volume),
        state_update=update_types.StateUpdate(
            pipette_aspirated_fluid=update_types.PipetteEjectedFluidUpdate(
                pipette_id="pipette-id-abc", volume=dispense_volume
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="pipette-id-abc", ready_to_aspirate=expected
            ),
        ),
    )


@pytest.mark.parametrize(
    "location,stateupdateLabware,stateupdateWell",
    [
        (
            CurrentWell(
                pipette_id="pipette-id",
                labware_id="labware-id-1",
                well_name="well-name-1",
            ),
            "labware-id-1",
            "well-name-1",
        ),
        (None, None, None),
        (CurrentAddressableArea("pipette-id", "addressable-area-1"), None, None),
    ],
)
async def test_overpressure_error(
    decoy: Decoy,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    state_view: StateView,
    model_utils: ModelUtils,
    subject: DispenseInPlaceImplementation,
    location: CurrentPipetteLocation | None,
    stateupdateLabware: str,
    stateupdateWell: str,
) -> None:
    """It should return an overpressure error if the hardware API indicates that."""
    pipette_id = "pipette-id"

    position = Point(x=1, y=2, z=3)

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    data = DispenseInPlaceParams(
        pipetteId=pipette_id,
        volume=50,
        flowRate=1.23,
        pushOut=10,
    )

    decoy.when(
        state_view.geometry.get_nozzles_per_well(
            labware_id=stateupdateLabware,
            target_well_name=stateupdateWell,
            pipette_id="pipette-id",
        )
    ).then_return(2)

    decoy.when(
        state_view.geometry.get_wells_covered_by_pipette_with_active_well(
            stateupdateLabware, stateupdateWell, "pipette-id"
        )
    ).then_return(["A3", "A4"])

    decoy.when(state_view.pipettes.get_aspirated_volume("pipette-id")).then_return(100)

    decoy.when(pipetting.get_state_view()).then_return(state_view)

    decoy.when(
        await pipetting.dispense_in_place(
            pipette_id=pipette_id,
            volume=50,
            flow_rate=1.23,
            push_out=10,
            is_full_dispense=False,
            correction_volume=0,
        ),
    ).then_raise(PipetteOverpressureError())

    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)
    decoy.when(await gantry_mover.get_position(pipette_id)).then_return(position)
    decoy.when(state_view.pipettes.get_current_location()).then_return(location)

    result = await subject.execute(data)

    if isinstance(location, CurrentWell):
        assert result == DefinedErrorData(
            public=OverpressureError.model_construct(
                id=error_id,
                createdAt=error_timestamp,
                wrappedErrors=[matchers.Anything()],
                errorInfo={"retryLocation": (position.x, position.y, position.z)},
            ),
            state_update=update_types.StateUpdate(
                liquid_operated=update_types.LiquidOperatedUpdate(
                    labware_id=stateupdateLabware,
                    well_names=["A3", "A4"],
                    volume_added=update_types.CLEAR,
                ),
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id", ready_to_aspirate=False
                ),
            ),
        )
    else:
        assert result == DefinedErrorData(
            public=OverpressureError.model_construct(
                id=error_id,
                createdAt=error_timestamp,
                wrappedErrors=[matchers.Anything()],
                errorInfo={"retryLocation": (position.x, position.y, position.z)},
            ),
            state_update=update_types.StateUpdate(
                pipette_aspirated_fluid=update_types.PipetteUnknownFluidUpdate(
                    pipette_id="pipette-id"
                ),
                ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                    pipette_id="pipette-id", ready_to_aspirate=False
                ),
            ),
        )



"""Test robot.move-to commands."""

from decoy import Decoy

from opentrons.protocol_engine.commands.command import SuccessData
from opentrons.protocol_engine.commands.robot.move_to import (
    MoveToImplementation,
    MoveToParams,
    MoveToResult,
)
from opentrons.protocol_engine.execution import MovementHandler
from opentrons.protocol_engine.types import DeckPoint
from opentrons.types import MountType, Point


async def test_move_to_implementation(
    decoy: Decoy,
    movement: MovementHandler,
) -> None:
    """Test the `robot.moveTo` implementation.

    It should call `MovementHandler.move_mount_to` with the
    correct coordinates.
    """
    subject = MoveToImplementation(
        movement=movement,
    )

    params = MoveToParams(
        mount=MountType.LEFT,
        destination=DeckPoint(x=1.11, y=2.22, z=3.33),
        speed=567.8,
    )

    decoy.when(
        await movement.move_mount_to(
            mount=MountType.LEFT,
            destination=DeckPoint(x=1.11, y=2.22, z=3.33),
            speed=567.8,
        )
    ).then_return(Point(x=4.44, y=5.55, z=6.66))

    result = await subject.execute(params=params)

    assert result == SuccessData(
        public=MoveToResult(position=DeckPoint(x=4.44, y=5.55, z=6.66))
    )



"""Test LiquidProbe commands."""

from datetime import datetime
from typing import Type, Union

import pytest
from decoy import Decoy, matchers

from opentrons_shared_data.errors.exceptions import (
    PipetteLiquidNotFoundError,
    StallOrCollisionDetectedError,
)
from opentrons_shared_data.pipette.pipette_definition import (
    AvailableSensorDefinition,
    SupportedTipsDefinition,
)
from opentrons_shared_data.pipette.types import (
    LiquidClasses as VolumeModes,
)
from opentrons_shared_data.pipette.types import (
    PipetteNameType,
)

from ..pipette_fixtures import get_default_nozzle_map
from opentrons.protocol_engine import DeckPoint, WellLocation, WellOffset, WellOrigin
from opentrons.protocol_engine.commands.command import DefinedErrorData, SuccessData
from opentrons.protocol_engine.commands.liquid_probe import (
    LiquidProbeImplementation,
    LiquidProbeParams,
    LiquidProbeResult,
    TryLiquidProbeImplementation,
    TryLiquidProbeParams,
    TryLiquidProbeResult,
)
from opentrons.protocol_engine.commands.movement_common import StallOrCollisionError
from opentrons.protocol_engine.commands.pipetting_common import LiquidNotFoundError
from opentrons.protocol_engine.errors.exceptions import (
    MustHomeError,
    PipetteNotReadyToAspirateError,
    TipNotAttachedError,
    TipNotEmptyError,
)
from opentrons.protocol_engine.execution import (
    GantryMover,
    MovementHandler,
    PipettingHandler,
)
from opentrons.protocol_engine.execution.pipetting import VirtualPipettingHandler
from opentrons.protocol_engine.resources.model_utils import ModelUtils
from opentrons.protocol_engine.state import update_types
from opentrons.protocol_engine.state.pipettes import (
    BoundingNozzlesOffsets,
    PipetteBoundingBoxOffsets,
    StaticPipetteConfig,
)
from opentrons.protocol_engine.state.state import StateView
from opentrons.protocol_engine.types import LabwareWellId
from opentrons.protocol_engine.types.liquid_level_detection import SimulatedProbeResult
from opentrons.types import MountType, Point

EitherImplementationType = Union[
    Type[LiquidProbeImplementation], Type[TryLiquidProbeImplementation]
]
EitherImplementation = Union[LiquidProbeImplementation, TryLiquidProbeImplementation]
EitherParamsType = Union[Type[LiquidProbeParams], Type[TryLiquidProbeParams]]
EitherResultType = Union[Type[LiquidProbeResult], Type[TryLiquidProbeResult]]


@pytest.fixture
def available_sensors() -> AvailableSensorDefinition:
    """Provide a list of sensors."""
    return AvailableSensorDefinition(sensors=["pressure", "capacitive", "environment"])


@pytest.fixture(
    params=[
        (LiquidProbeImplementation, LiquidProbeParams, LiquidProbeResult),
        (TryLiquidProbeImplementation, TryLiquidProbeParams, TryLiquidProbeResult),
    ]
)
def types(
    request: pytest.FixtureRequest,
) -> tuple[EitherImplementationType, EitherParamsType, EitherResultType]:
    """Return a tuple of types associated with a single variant of the command."""
    return request.param  # type: ignore[no-any-return]


@pytest.fixture
def implementation_type(
    types: tuple[EitherImplementationType, object, object],
) -> EitherImplementationType:
    """Return an implementation type. Kept in sync with the params and result types."""
    return types[0]


@pytest.fixture
def params_type(types: tuple[object, EitherParamsType, object]) -> EitherParamsType:
    """Return a params type. Kept in sync with the implementation and result types."""
    return types[1]


@pytest.fixture
def result_type(types: tuple[object, object, EitherResultType]) -> EitherResultType:
    """Return a result type. Kept in sync with the implementation and params types."""
    return types[2]


@pytest.fixture
def subject(
    implementation_type: EitherImplementationType,
    state_view: StateView,
    movement: MovementHandler,
    gantry_mover: GantryMover,
    pipetting: PipettingHandler,
    model_utils: ModelUtils,
) -> Union[LiquidProbeImplementation, TryLiquidProbeImplementation]:
    """Get the implementation subject."""
    return implementation_type(
        state_view=state_view,
        pipetting=pipetting,
        movement=movement,
        gantry_mover=gantry_mover,
        model_utils=model_utils,
    )


async def test_virtual_liquid_probe(
    decoy: Decoy,
) -> None:
    """Check that VirtualPipettingHandler::liquid_probe_in_place returns a SimulatedProbeResult."""
    mock_state_view = decoy.mock(cls=StateView)
    subject = VirtualPipettingHandler(state_view=mock_state_view)

    location = WellLocation(origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1))

    liquid_probe_result = await subject.liquid_probe_in_place(
        pipette_id="abc",
        labware_id="123",
        well_name="A3",
        well_location=location,
    )
    assert isinstance(liquid_probe_result, SimulatedProbeResult)


async def test_liquid_probe_implementation(
    decoy: Decoy,
    movement: MovementHandler,
    state_view: StateView,
    pipetting: PipettingHandler,
    subject: EitherImplementation,
    params_type: EitherParamsType,
    result_type: EitherResultType,
    model_utils: ModelUtils,
    available_sensors: AvailableSensorDefinition,
    supported_tip_fixture: SupportedTipsDefinition,
) -> None:
    """It should move to the destination and do a liquid probe there."""
    location = WellLocation(origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1))

    data = params_type(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=location,
    )

    decoy.when(state_view.pipettes.get_aspirated_volume(pipette_id="abc")).then_return(
        0
    )

    decoy.when(
        await movement.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(Point(x=1, y=2, z=3))

    decoy.when(
        await pipetting.liquid_probe_in_place(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=location,
        ),
    ).then_return(15.0)

    decoy.when(
        state_view.geometry.get_well_volume_at_height(
            labware_id="123",
            well_name="A3",
            height=15.0,
        ),
    ).then_return(30.0)
    decoy.when(
        state_view.pipettes.get_nozzle_configuration_supports_lld("abc")
    ).then_return(True)

    decoy.when(state_view.pipettes.get_config("abc")).then_return(
        StaticPipetteConfig(
            min_volume=1,
            max_volume=9001,
            channels=1,
            model="blah",
            display_name="bleh",
            serial_number="",
            tip_configuration_lookup_table={9001: supported_tip_fixture},
            nominal_tip_overlap={},
            home_position=0,
            nozzle_offset_z=0,
            bounding_nozzle_offsets=BoundingNozzlesOffsets(
                back_left_offset=Point(x=10, y=20, z=30),
                front_right_offset=Point(x=40, y=50, z=60),
            ),
            default_nozzle_map=get_default_nozzle_map(PipetteNameType.P1000_96),
            pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                back_left_corner=Point(x=10, y=20, z=30),
                front_right_corner=Point(x=40, y=50, z=60),
                front_left_corner=Point(x=10, y=50, z=60),
                back_right_corner=Point(x=40, y=20, z=60),
            ),
            lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=available_sensors,
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        )
    )

    timestamp = datetime(year=2020, month=1, day=2)
    decoy.when(model_utils.get_timestamp()).then_return(timestamp)

    result = await subject.execute(data)

    assert type(result.public) is result_type  # Pydantic v1 only compares the fields.
    assert result == SuccessData(
        public=result_type(z_position=15.0, position=DeckPoint(x=1, y=2, z=3)),
        state_update=update_types.StateUpdate(
            pipette_location=update_types.PipetteLocationUpdate(
                pipette_id="abc",
                new_location=LabwareWellId(labware_id="123", well_name="A3"),
                new_deck_point=DeckPoint(x=1, y=2, z=3),
            ),
            liquid_probed=update_types.LiquidProbedUpdate(
                labware_id="123",
                well_name="A3",
                height=15.0,
                volume=30.0,
                last_probed=timestamp,
            ),
            ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
                pipette_id="abc", ready_to_aspirate=True
            ),
        ),
    )


async def test_liquid_not_found_error(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    pipetting: PipettingHandler,
    subject: EitherImplementation,
    params_type: EitherParamsType,
    model_utils: ModelUtils,
    available_sensors: AvailableSensorDefinition,
    supported_tip_fixture: SupportedTipsDefinition,
) -> None:
    """It should return a liquid not found error if the hardware API indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = WellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    position = Point(x=1, y=2, z=3)

    error_id = "error-id"
    error_timestamp = datetime(year=2020, month=1, day=2)

    data = params_type(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
    )

    decoy.when(state_view.pipettes.get_aspirated_volume(pipette_id)).then_return(0)
    decoy.when(state_view.pipettes.get_config("pipette-id")).then_return(
        StaticPipetteConfig(
            min_volume=1,
            max_volume=9001,
            channels=1,
            model="blah",
            display_name="bleh",
            serial_number="",
            tip_configuration_lookup_table={9001: supported_tip_fixture},
            nominal_tip_overlap={},
            home_position=0,
            nozzle_offset_z=0,
            bounding_nozzle_offsets=BoundingNozzlesOffsets(
                back_left_offset=Point(x=10, y=20, z=30),
                front_right_offset=Point(x=40, y=50, z=60),
            ),
            default_nozzle_map=get_default_nozzle_map(PipetteNameType.P1000_96),
            pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                back_left_corner=Point(x=10, y=20, z=30),
                front_right_corner=Point(x=40, y=50, z=60),
                front_left_corner=Point(x=10, y=50, z=60),
                back_right_corner=Point(x=40, y=20, z=60),
            ),
            lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=available_sensors,
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        )
    )
    decoy.when(
        await movement.move_to_well(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_return(position)
    decoy.when(
        await pipetting.liquid_probe_in_place(
            pipette_id=pipette_id,
            labware_id=labware_id,
            well_name=well_name,
            well_location=well_location,
        ),
    ).then_raise(PipetteLiquidNotFoundError())
    decoy.when(
        state_view.pipettes.get_nozzle_configuration_supports_lld(pipette_id)
    ).then_return(True)
    decoy.when(model_utils.generate_id()).then_return(error_id)
    decoy.when(model_utils.get_timestamp()).then_return(error_timestamp)

    result = await subject.execute(data)

    expected_state_update = update_types.StateUpdate(
        pipette_location=update_types.PipetteLocationUpdate(
            pipette_id=pipette_id,
            new_location=LabwareWellId(labware_id=labware_id, well_name=well_name),
            new_deck_point=DeckPoint(x=position.x, y=position.y, z=position.z),
        ),
        liquid_probed=update_types.LiquidProbedUpdate(
            labware_id=labware_id,
            well_name=well_name,
            height=update_types.CLEAR,
            volume=update_types.CLEAR,
            last_probed=error_timestamp,
        ),
        ready_to_aspirate=update_types.PipetteAspirateReadyUpdate(
            pipette_id=pipette_id, ready_to_aspirate=True
        ),
    )
    if isinstance(subject, LiquidProbeImplementation):
        assert result == DefinedErrorData(
            public=LiquidNotFoundError.model_construct(
                id=error_id,
                createdAt=error_timestamp,
                wrappedErrors=[matchers.Anything()],
            ),
            state_update=expected_state_update,
        )
    else:
        assert result == SuccessData(
            public=TryLiquidProbeResult(
                z_position=None,
                position=DeckPoint(x=position.x, y=position.y, z=position.z),
            ),
            state_update=expected_state_update,
        )


async def test_liquid_probe_tip_checking(
    decoy: Decoy,
    state_view: StateView,
    subject: EitherImplementation,
    params_type: EitherParamsType,
    available_sensors: AvailableSensorDefinition,
    supported_tip_fixture: SupportedTipsDefinition,
) -> None:
    """It should raise a TipNotAttached error if the state view indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = WellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    data = params_type(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
    )
    decoy.when(
        state_view.pipettes.get_nozzle_configuration_supports_lld(pipette_id)
    ).then_return(True)
    decoy.when(state_view.pipettes.get_aspirated_volume(pipette_id)).then_raise(
        TipNotAttachedError()
    )
    decoy.when(state_view.pipettes.get_config("pipette-id")).then_return(
        StaticPipetteConfig(
            min_volume=1,
            max_volume=9001,
            channels=1,
            model="blah",
            display_name="bleh",
            serial_number="",
            tip_configuration_lookup_table={9001: supported_tip_fixture},
            nominal_tip_overlap={},
            home_position=0,
            nozzle_offset_z=0,
            bounding_nozzle_offsets=BoundingNozzlesOffsets(
                back_left_offset=Point(x=10, y=20, z=30),
                front_right_offset=Point(x=40, y=50, z=60),
            ),
            default_nozzle_map=get_default_nozzle_map(PipetteNameType.P1000_96),
            pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                back_left_corner=Point(x=10, y=20, z=30),
                front_right_corner=Point(x=40, y=50, z=60),
                front_left_corner=Point(x=10, y=50, z=60),
                back_right_corner=Point(x=40, y=20, z=60),
            ),
            lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=available_sensors,
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        )
    )
    with pytest.raises(TipNotAttachedError):
        await subject.execute(data)


async def test_liquid_probe_plunger_preparedness_checking(
    decoy: Decoy,
    state_view: StateView,
    subject: EitherImplementation,
    params_type: EitherParamsType,
    available_sensors: AvailableSensorDefinition,
    supported_tip_fixture: SupportedTipsDefinition,
) -> None:
    """It should raise a PipetteNotReadyToAspirate error if the state view indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = WellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    data = params_type(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
    )
    decoy.when(
        state_view.pipettes.get_nozzle_configuration_supports_lld(pipette_id)
    ).then_return(True)
    decoy.when(state_view.pipettes.get_config("pipette-id")).then_return(
        StaticPipetteConfig(
            min_volume=1,
            max_volume=9001,
            channels=1,
            model="blah",
            display_name="bleh",
            serial_number="",
            tip_configuration_lookup_table={9001: supported_tip_fixture},
            nominal_tip_overlap={},
            home_position=0,
            nozzle_offset_z=0,
            bounding_nozzle_offsets=BoundingNozzlesOffsets(
                back_left_offset=Point(x=10, y=20, z=30),
                front_right_offset=Point(x=40, y=50, z=60),
            ),
            default_nozzle_map=get_default_nozzle_map(PipetteNameType.P1000_96),
            pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                back_left_corner=Point(x=10, y=20, z=30),
                front_right_corner=Point(x=40, y=50, z=60),
                front_left_corner=Point(x=10, y=50, z=60),
                back_right_corner=Point(x=40, y=20, z=60),
            ),
            lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=available_sensors,
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        )
    )
    decoy.when(state_view.pipettes.get_aspirated_volume(pipette_id)).then_return(None)
    with pytest.raises(PipetteNotReadyToAspirateError):
        await subject.execute(data)


async def test_liquid_probe_volume_checking(
    decoy: Decoy,
    state_view: StateView,
    subject: EitherImplementation,
    params_type: EitherParamsType,
    available_sensors: AvailableSensorDefinition,
    supported_tip_fixture: SupportedTipsDefinition,
) -> None:
    """It should return a TipNotEmptyError if the hardware API indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = WellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    data = params_type(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
    )

    decoy.when(
        state_view.pipettes.get_aspirated_volume(pipette_id=pipette_id),
    ).then_return(123)
    decoy.when(state_view.pipettes.get_config("pipette-id")).then_return(
        StaticPipetteConfig(
            min_volume=1,
            max_volume=9001,
            channels=1,
            model="blah",
            display_name="bleh",
            serial_number="",
            tip_configuration_lookup_table={9001: supported_tip_fixture},
            nominal_tip_overlap={},
            home_position=0,
            nozzle_offset_z=0,
            bounding_nozzle_offsets=BoundingNozzlesOffsets(
                back_left_offset=Point(x=10, y=20, z=30),
                front_right_offset=Point(x=40, y=50, z=60),
            ),
            default_nozzle_map=get_default_nozzle_map(PipetteNameType.P1000_96),
            pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                back_left_corner=Point(x=10, y=20, z=30),
                front_right_corner=Point(x=40, y=50, z=60),
                front_left_corner=Point(x=10, y=50, z=60),
                back_right_corner=Point(x=40, y=20, z=60),
            ),
            lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=available_sensors,
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        )
    )
    decoy.when(
        state_view.pipettes.get_nozzle_configuration_supports_lld(pipette_id)
    ).then_return(True)

    with pytest.raises(TipNotEmptyError):
        await subject.execute(data)

    decoy.when(
        state_view.pipettes.get_aspirated_volume(pipette_id=pipette_id),
    ).then_return(None)

    with pytest.raises(PipetteNotReadyToAspirateError):
        await subject.execute(data)


async def test_liquid_probe_location_checking(
    decoy: Decoy,
    state_view: StateView,
    movement: MovementHandler,
    subject: EitherImplementation,
    params_type: EitherParamsType,
    available_sensors: AvailableSensorDefinition,
    supported_tip_fixture: SupportedTipsDefinition,
) -> None:
    """It should return a PositionUnkownError if the hardware API indicates that."""
    pipette_id = "pipette-id"
    labware_id = "labware-id"
    well_name = "well-name"
    well_location = WellLocation(
        origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1)
    )

    data = params_type(
        pipetteId=pipette_id,
        labwareId=labware_id,
        wellName=well_name,
        wellLocation=well_location,
    )
    decoy.when(state_view.pipettes.get_aspirated_volume(pipette_id)).then_return(0)
    decoy.when(state_view.pipettes.get_config("pipette-id")).then_return(
        StaticPipetteConfig(
            min_volume=1,
            max_volume=9001,
            channels=1,
            model="blah",
            display_name="bleh",
            serial_number="",
            tip_configuration_lookup_table={9001: supported_tip_fixture},
            nominal_tip_overlap={},
            home_position=0,
            nozzle_offset_z=0,
            bounding_nozzle_offsets=BoundingNozzlesOffsets(
                back_left_offset=Point(x=10, y=20, z=30),
                front_right_offset=Point(x=40, y=50, z=60),
            ),
            default_nozzle_map=get_default_nozzle_map(PipetteNameType.P1000_96),
            pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                back_left_corner=Point(x=10, y=20, z=30),
                front_right_corner=Point(x=40, y=50, z=60),
                front_left_corner=Point(x=10, y=50, z=60),
                back_right_corner=Point(x=40, y=20, z=60),
            ),
            lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=available_sensors,
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        )
    )
    decoy.when(
        await movement.check_for_valid_position(
            mount=MountType.LEFT,
        ),
    ).then_return(False)
    decoy.when(
        state_view.pipettes.get_nozzle_configuration_supports_lld(pipette_id)
    ).then_return(True)
    with pytest.raises(MustHomeError):
        await subject.execute(data)


async def test_liquid_probe_stall(
    decoy: Decoy,
    movement: MovementHandler,
    state_view: StateView,
    pipetting: PipettingHandler,
    subject: EitherImplementation,
    params_type: EitherParamsType,
    model_utils: ModelUtils,
    available_sensors: AvailableSensorDefinition,
    supported_tip_fixture: SupportedTipsDefinition,
) -> None:
    """It should move to the destination and do a liquid probe there."""
    location = WellLocation(origin=WellOrigin.BOTTOM, offset=WellOffset(x=0, y=0, z=1))

    data = params_type(
        pipetteId="abc",
        labwareId="123",
        wellName="A3",
        wellLocation=location,
    )

    decoy.when(state_view.pipettes.get_aspirated_volume(pipette_id="abc")).then_return(
        0
    )
    decoy.when(state_view.pipettes.get_config("abc")).then_return(
        StaticPipetteConfig(
            min_volume=1,
            max_volume=9001,
            channels=1,
            model="blah",
            display_name="bleh",
            serial_number="",
            tip_configuration_lookup_table={9001: supported_tip_fixture},
            nominal_tip_overlap={},
            home_position=0,
            nozzle_offset_z=0,
            bounding_nozzle_offsets=BoundingNozzlesOffsets(
                back_left_offset=Point(x=10, y=20, z=30),
                front_right_offset=Point(x=40, y=50, z=60),
            ),
            default_nozzle_map=get_default_nozzle_map(PipetteNameType.P1000_96),
            pipette_bounding_box_offsets=PipetteBoundingBoxOffsets(
                back_left_corner=Point(x=10, y=20, z=30),
                front_right_corner=Point(x=40, y=50, z=60),
                front_left_corner=Point(x=10, y=50, z=60),
                back_right_corner=Point(x=40, y=20, z=60),
            ),
            lld_settings={},
            plunger_positions={
                "top": 0.0,
                "bottom": 5.0,
                "blow_out": 19.0,
                "drop_tip": 20.0,
            },
            shaft_ul_per_mm=5.0,
            available_sensors=available_sensors,
            volume_mode=VolumeModes.default,
            available_volume_modes_min_vol={},
        )
    )
    decoy.when(
        state_view.pipettes.get_nozzle_configuration_supports_lld("abc")
    ).then_return(True)

    decoy.when(
        await movement.move_to_well(
            pipette_id="abc",
            labware_id="123",
            well_name="A3",
            well_location=location,
            current_well=None,
            force_direct=False,
            minimum_z_height=None,
            speed=None,
            operation_volume=None,
            offset_pipette_for_reservoir_subwells=False,
        ),
    ).then_raise(StallOrCollisionDetectedError())

    error_id = "error-id"
    timestamp = datetime(year=2020, month=1, day=2)
    decoy.when(model_utils.get_timestamp()).then_return(timestamp)
    decoy.when(model_utils.generate_id()).then_return(error_id)

    result = await subject.execute(data)

    assert result == DefinedErrorData(
        public=StallOrCollisionError.model_construct(
            id=error_id, createdAt=timestamp, wrappedErrors=[matchers.Anything()]
        ),
        state_update=update_types.StateUpdate(pipette_location=update_types.CLEAR),
    )



"""Tests for the engine based Protocol API module core implementations."""

import pytest
from decoy import Decoy

from opentrons.drivers.types import HeaterShakerLabwareLatchStatus
from opentrons.hardware_control import SynchronousAdapter
from opentrons.hardware_control.modules import HeaterShaker
from opentrons.hardware_control.modules.types import (
    ModuleType,
    SpeedStatus,
    TemperatureStatus,
)
from opentrons.protocol_api import MAX_SUPPORTED_VERSION
from opentrons.protocol_api.core.engine.module_core import HeaterShakerModuleCore
from opentrons.protocol_api.core.engine.protocol import ProtocolCore
from opentrons.protocol_api.core.engine.tasks import EngineTaskCore
from opentrons.protocol_engine import commands as cmd
from opentrons.protocol_engine.clients import SyncClient as EngineClient

SyncHeaterShakerHardware = SynchronousAdapter[HeaterShaker]


@pytest.fixture
def mock_engine_client(decoy: Decoy) -> EngineClient:
    """Get a mock ProtocolEngine synchronous client."""
    return decoy.mock(cls=EngineClient)


@pytest.fixture
def mock_sync_module_hardware(decoy: Decoy) -> SyncHeaterShakerHardware:
    """Get a mock synchronous module hardware."""
    return decoy.mock(name="SyncHeaterShakerHardware")  # type: ignore[no-any-return]


@pytest.fixture
def mock_protocol_core(decoy: Decoy) -> ProtocolCore:
    """Get a mock protocol core."""
    return decoy.mock(cls=ProtocolCore)


@pytest.fixture
def subject(
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    mock_protocol_core: ProtocolCore,
) -> HeaterShakerModuleCore:
    """Get a HeaterShakerModuleCore test subject."""
    return HeaterShakerModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )


def test_create(
    decoy: Decoy,
    mock_engine_client: EngineClient,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    mock_protocol_core: ProtocolCore,
) -> None:
    """It should be able to create a heater shaker module core."""
    result = HeaterShakerModuleCore(
        module_id="1234",
        engine_client=mock_engine_client,
        api_version=MAX_SUPPORTED_VERSION,
        sync_module_hardware=mock_sync_module_hardware,
        protocol_core=mock_protocol_core,
    )

    assert result.module_id == "1234"
    assert result.MODULE_TYPE == ModuleType.HEATER_SHAKER


def test_set_target_temperature(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should set the target temperature with the engine client and return a EngineTaskCore."""
    task_mock = decoy.mock(cls=EngineTaskCore)
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.heater_shaker.SetTargetTemperatureParams(moduleId="1234", celsius=42.0)
        )
    ).then_return(cmd.heater_shaker.SetTargetTemperatureResult(taskId="taskId"))
    task_mock._id = "taskId"
    result = subject.set_target_temperature(42.0)
    assert isinstance(result, EngineTaskCore)
    assert result._id == "taskId"


def test_wait_for_target_temperature(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should wait for the target temperature with the engine client."""
    subject.wait_for_target_temperature()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.heater_shaker.WaitForTemperatureParams(moduleId="1234")
        ),
        times=1,
    )


def test_set_and_wait_for_shake_speed(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should set and wait for shake speed with the engine client."""
    subject.set_and_wait_for_shake_speed(rpm=1337)

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.heater_shaker.SetAndWaitForShakeSpeedParams(moduleId="1234", rpm=1337)
        ),
        times=1,
    )


def test_shake_speed(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should set and wait for shake speed with the engine client."""
    task_mock = decoy.mock(cls=EngineTaskCore)
    decoy.when(
        mock_engine_client.execute_command_without_recovery(
            cmd.heater_shaker.SetShakeSpeedParams(moduleId="1234", rpm=1337)
        )
    ).then_return(
        cmd.heater_shaker.SetShakeSpeedResult(taskId="taskId", pipetteRetracted=True)
    )
    task_mock._id = "taskId"
    result = subject.set_shake_speed(1337)
    assert isinstance(result, EngineTaskCore)
    assert result._id == "taskId"


def test_open_labware_latch(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should open the labware latch with the engine client."""
    subject.open_labware_latch()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.heater_shaker.OpenLabwareLatchParams(moduleId="1234")
        ),
        times=1,
    )


def test_close_labware_latch(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should close the labware latch with the engine client."""
    subject.close_labware_latch()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.heater_shaker.CloseLabwareLatchParams(moduleId="1234")
        ),
        times=1,
    )


def test_deactivate_shaker(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should deactivate the shaker with the engine client."""
    subject.deactivate_shaker()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.heater_shaker.DeactivateShakerParams(moduleId="1234")
        ),
        times=1,
    )


def test_deactivate_heater(
    decoy: Decoy, mock_engine_client: EngineClient, subject: HeaterShakerModuleCore
) -> None:
    """It should deactivate the heater with the engine client."""
    subject.deactivate_heater()

    decoy.verify(
        mock_engine_client.execute_command(
            cmd.heater_shaker.DeactivateHeaterParams(moduleId="1234")
        ),
        times=1,
    )


def test_get_current_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: HeaterShakerModuleCore,
) -> None:
    """It should report the current temperature."""
    decoy.when(mock_sync_module_hardware.temperature).then_return(42.0)
    result = subject.get_current_temperature()
    assert result == 42.0


def test_get_target_temperature(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: HeaterShakerModuleCore,
) -> None:
    """It should report the target temperature."""
    decoy.when(mock_sync_module_hardware.target_temperature).then_return(42.0)
    result = subject.get_target_temperature()
    assert result == 42.0


def test_get_temperature_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: HeaterShakerModuleCore,
) -> None:
    """It should report the temperature status."""
    decoy.when(mock_sync_module_hardware.temperature_status).then_return(
        TemperatureStatus.COOLING
    )
    result = subject.get_temperature_status()
    assert result == TemperatureStatus.COOLING


def test_get_current_speed(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: HeaterShakerModuleCore,
) -> None:
    """It should report the current speed."""
    decoy.when(mock_sync_module_hardware.speed).then_return(321)
    result = subject.get_current_speed()
    assert result == 321


def test_get_target_speed(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: HeaterShakerModuleCore,
) -> None:
    """It should report the target speed."""
    decoy.when(mock_sync_module_hardware.target_speed).then_return(321)
    result = subject.get_target_speed()
    assert result == 321


def test_get_speed_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: HeaterShakerModuleCore,
) -> None:
    """It should report the speed status."""
    decoy.when(mock_sync_module_hardware.speed_status).then_return(
        SpeedStatus.ACCELERATING
    )
    result = subject.get_speed_status()
    assert result == SpeedStatus.ACCELERATING


def test_get_labware_latch_status(
    decoy: Decoy,
    mock_sync_module_hardware: SyncHeaterShakerHardware,
    subject: HeaterShakerModuleCore,
) -> None:
    """It should report the labware latch status."""
    decoy.when(mock_sync_module_hardware.labware_latch_status).then_return(
        HeaterShakerLabwareLatchStatus.OPENING
    )
    result = subject.get_labware_latch_status()
    assert result == HeaterShakerLabwareLatchStatus.OPENING


def test_get_serial_number(
    decoy: Decoy, subject: HeaterShakerModuleCore, mock_engine_client: EngineClient
) -> None:
    """It should return a serial number."""
    decoy.when(mock_engine_client.state.modules.get_serial_number("1234")).then_return(
        "abc"
    )

    assert subject.get_serial_number() == "abc"
